# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy

import mock
import six

from senlin.common import exception as exc
from senlin.objects import node as node_obj
from senlin.profiles.os.nova import server
from senlin.tests.unit.common import base
from senlin.tests.unit.common import utils


class TestServerNameChecking(base.SenlinTestCase):

    scenarios = [
        ('none-none', dict(
            old_name=None,
            new_name=None,
            result=(False, 'NODE_NAME'))),
        ('none-new', dict(
            old_name=None,
            new_name='NEW_NAME',
            result=(True, 'NEW_NAME'))),
        ('old-none', dict(
            old_name='OLD_NAME',
            new_name=None,
            result=(True, 'NODE_NAME'))),
        ('old-new', dict(
            old_name='OLD_NAME',
            new_name='NEW_NAME',
            result=(True, 'NEW_NAME')))
    ]

    def setUp(self):
        super(TestServerNameChecking, self).setUp()
        self.old_spec = {
            'type': 'os.nova.server',
            'version': '1.0',
            'properties': {
                'flavor': 'FLAVOR',
            }
        }
        self.new_spec = copy.deepcopy(self.old_spec)
        obj = mock.Mock()
        obj.name = 'NODE_NAME'
        self.obj = obj

    def test_check_server_name(self):
        if self.old_name:
            self.old_spec['properties']['name'] = self.old_name
        if self.new_name:
            self.new_spec['properties']['name'] = self.new_name

        profile = server.ServerProfile('t', self.old_spec)
        new_profile = server.ServerProfile('t1', self.new_spec)

        res = profile._check_server_name(self.obj, new_profile)

        self.assertEqual(self.result, res)


class TestPasswordChecking(base.SenlinTestCase):

    scenarios = [
        ('none-none', dict(
            old_passwd=None,
            new_passwd=None,
            result=(False, ''))),
        ('none-new', dict(
            old_passwd=None,
            new_passwd='NEW_PASSWD',
            result=(True, 'NEW_PASSWD'))),
        ('old-none', dict(
            old_passwd='OLD_PASSWD',
            new_passwd=None,
            result=(True, ''))),
        ('old-new', dict(
            old_passwd='OLD_PASSWD',
            new_passwd='NEW_PASSWD',
            result=(True, 'NEW_PASSWD')))
    ]

    def setUp(self):
        super(TestPasswordChecking, self).setUp()
        self.old_spec = {
            'type': 'os.nova.server',
            'version': '1.0',
            'properties': {
                'flavor': 'FLAVOR',
            }
        }
        self.new_spec = copy.deepcopy(self.old_spec)
        self.obj = mock.Mock()

    def test_check_password(self):
        if self.old_passwd:
            self.old_spec['properties']['admin_pass'] = self.old_passwd
        if self.new_passwd:
            self.new_spec['properties']['admin_pass'] = self.new_passwd

        profile = server.ServerProfile('t', self.old_spec)
        new_profile = server.ServerProfile('t1', self.new_spec)

        res = profile._check_password(self.obj, new_profile)

        self.assertEqual(self.result, res)


class TestNovaServerUpdate(base.SenlinTestCase):
    def setUp(self):
        super(TestNovaServerUpdate, self).setUp()

        self.context = utils.dummy_context()
        self.spec = {
            'type': 'os.nova.server',
            'version': '1.0',
            'properties': {
                'context': {},
                'admin_pass': 'adminpass',
                'auto_disk_config': True,
                'availability_zone': 'FAKE_AZ',
                'block_device_mapping': [{
                    'device_name': 'FAKE_NAME',
                    'volume_size': 1000,
                }],
                'config_drive': False,
                'flavor': 'FLAV',
                'image': 'FAKE_IMAGE',
                'key_name': 'FAKE_KEYNAME',
                "metadata": {"meta var": "meta val"},
                'name': 'FAKE_SERVER_NAME',
                'networks': [{
                    'port': 'FAKE_PORT',
                    'fixed_ip': 'FAKE_IP',
                    'network': 'FAKE_NET',
                }],
                'personality': [{
                    'path': '/etc/motd',
                    'contents': 'foo',
                }],
                'scheduler_hints': {
                    'same_host': 'HOST_ID',
                },
                'security_groups': ['HIGH_SECURITY_GROUP'],
                'user_data': 'FAKE_USER_DATA',
            }
        }
        self.patchobject(node_obj.Node, 'update')

    def test__update_name(self):
        profile = server.ServerProfile('t', self.spec)
        cc = mock.Mock()
        profile._computeclient = cc
        obj = mock.Mock(physical_id='NOVA_ID')

        res = profile._update_name(obj, 'NEW_NAME')

        self.assertIsNone(res)
        cc.server_update.assert_called_once_with('NOVA_ID', name='NEW_NAME')

    def test__update_name_nova_failure(self):
        profile = server.ServerProfile('t', self.spec)
        cc = mock.Mock()
        profile._computeclient = cc
        cc.server_update.side_effect = exc.InternalError(message='BOOM')
        obj = mock.Mock(physical_id='NOVA_ID')

        ex = self.assertRaises(exc.EResourceUpdate,
                               profile._update_name,
                               obj, 'NEW_NAME')

        self.assertEqual("Failed in updating server 'NOVA_ID': BOOM.",
                         six.text_type(ex))
        cc.server_update.assert_called_once_with('NOVA_ID', name='NEW_NAME')

    def test__update_password(self):
        profile = server.ServerProfile('t', self.spec)
        cc = mock.Mock()
        profile._computeclient = cc
        obj = mock.Mock(physical_id='NOVA_ID')

        res = profile._update_password(obj, 'NEW_PASSWORD')

        self.assertIsNone(res)
        cc.server_change_password.assert_called_once_with(
            'NOVA_ID', 'NEW_PASSWORD')

    def test__update_password_nova_failure(self):
        profile = server.ServerProfile('t', self.spec)
        cc = mock.Mock()
        profile._computeclient = cc
        err = exc.InternalError(message='BOOM')
        cc.server_change_password.side_effect = err
        obj = mock.Mock(physical_id='NOVA_ID')

        ex = self.assertRaises(exc.EResourceUpdate,
                               profile._update_password,
                               obj, 'NEW_PASSWORD')

        self.assertEqual("Failed in updating server 'NOVA_ID': BOOM.",
                         six.text_type(ex))
        cc.server_change_password.assert_called_once_with(
            'NOVA_ID', 'NEW_PASSWORD')

    def test__update_metadata(self):
        obj = mock.Mock(id='NODE_ID', physical_id='NOVA_ID',
                        cluster_id='CLUSTER_ID', index=456)
        cc = mock.Mock()
        profile = server.ServerProfile('t', self.spec)
        profile._computeclient = cc
        new_spec = copy.deepcopy(self.spec)
        new_spec['properties']['metadata'] = {'new_key': 'new_value'}
        new_profile = server.ServerProfile('t', new_spec)

        res = profile._update_metadata(obj, new_profile)

        self.assertIsNone(res)
        cc.server_metadata_update.assert_called_once_with(
            'NOVA_ID',
            {
                'new_key': 'new_value',
                'cluster_node_id': 'NODE_ID',
                'cluster_id': 'CLUSTER_ID',
                'cluster_node_index': '456',
            }
        )

    def test___update_metadata_no_change(self):
        obj = mock.Mock(id='NODE_ID')
        profile = server.ServerProfile('t', self.spec)
        cc = mock.Mock()
        profile._computeclient = cc
        new_spec = copy.deepcopy(self.spec)
        new_profile = server.ServerProfile('t', new_spec)

        res = profile._update_metadata(obj, new_profile)

        self.assertIsNone(res)
        self.assertEqual(0, cc.server_metadata_update.call_count)

    def test__update_metadata_nova_failure(self):
        obj = mock.Mock(id='NODE_ID', physical_id='NOVA_ID', cluster_id='')
        err = exc.InternalError(code=500, message='Nova Error')
        cc = mock.Mock()
        cc.server_metadata_update.side_effect = err
        profile = server.ServerProfile('t', self.spec)
        profile._computeclient = cc

        # new profile with new metadata
        new_spec = copy.deepcopy(self.spec)
        new_spec['properties']['metadata'] = {'fooa': 'baaar'}
        new_profile = server.ServerProfile('t', new_spec)

        ex = self.assertRaises(exc.EResourceUpdate,
                               profile._update_metadata,
                               obj, new_profile)

        self.assertEqual("Failed in updating server 'NOVA_ID': "
                         "Nova Error.", six.text_type(ex))
        cc.server_metadata_update.assert_called_once_with(
            'NOVA_ID', {'fooa': 'baaar', 'cluster_node_id': 'NODE_ID'}
        )

    def test__update_flavor(self):
        obj = mock.Mock(physical_id='NOVA_ID')
        cc = mock.Mock()
        profile = server.ServerProfile('t', self.spec)
        profile._computeclient = cc
        x_flavors = [mock.Mock(id='123'), mock.Mock(id='456')]
        mock_validate = self.patchobject(profile, '_validate_flavor',
                                         side_effect=x_flavors)
        new_spec = copy.deepcopy(self.spec)
        new_spec['properties']['flavor'] = 'new_flavor'
        new_profile = server.ServerProfile('t1', new_spec)
        profile._update_flavor(obj, new_profile)

        mock_validate.assert_has_calls([
            mock.call(obj, 'FLAV', 'update'),
            mock.call(obj, 'new_flavor', 'update')
        ])
        cc.server_resize.assert_called_once_with('NOVA_ID', '456')
        cc.server_resize_confirm.assert_called_once_with('NOVA_ID')
        cc.wait_for_server.has_calls([
            mock.call('NOVA_ID', 'VERIFY_RESIZE'),
            mock.call('NOVA_ID', 'ACTIVE')])

    def test__update_flavor_failed_validation(self):
        obj = mock.Mock(physical_id='NOVA_ID')
        cc = mock.Mock()
        profile = server.ServerProfile('t', self.spec)
        profile._computeclient = cc
        new_spec = copy.deepcopy(self.spec)
        new_spec['properties']['flavor'] = 'new_flavor'
        new_profile = server.ServerProfile('t1', new_spec)
        err = exc.EResourceUpdate(type='server', id='NOVA_ID', message='BOOM')
        mock_validate = self.patchobject(profile, '_validate_flavor',
                                         side_effect=err)

        self.assertRaises(exc.EResourceUpdate,
                          profile._update_flavor,
                          obj, new_profile)

        mock_validate.assert_called_once_with(obj, 'FLAV', 'update')

    def test__update_flavor_failed_validation_2(self):
        obj = mock.Mock(physical_id='NOVA_ID')
        cc = mock.Mock()
        profile = server.ServerProfile('t', self.spec)
        profile._computeclient = cc
        new_spec = copy.deepcopy(self.spec)
        new_spec['properties']['flavor'] = 'new_flavor'
        new_profile = server.ServerProfile('t1', new_spec)
        result = [
            mock.Mock(),
            exc.EResourceUpdate(type='server', id='NOVA_ID', message='BOOM')
        ]
        mock_validate = self.patchobject(profile, '_validate_flavor',
                                         side_effect=result)

        self.assertRaises(exc.EResourceUpdate,
                          profile._update_flavor,
                          obj, new_profile)

        mock_validate.assert_has_calls([
            mock.call(obj, 'FLAV', 'update'),
            mock.call(obj, 'new_flavor', 'update'),
        ])

    def test__update_flavor_same(self):
        obj = mock.Mock(physical_id='NOVA_ID')
        cc = mock.Mock()
        profile = server.ServerProfile('t', self.spec)
        profile._computeclient = cc
        new_spec = copy.deepcopy(self.spec)
        new_profile = server.ServerProfile('t1', new_spec)

        x_flavors = [mock.Mock(id=123), mock.Mock(id=123)]
        mock_validate = self.patchobject(profile, '_validate_flavor',
                                         side_effect=x_flavors)

        res = profile._update_flavor(obj, new_profile)

        self.assertIsNone(res)
        mock_validate.assert_has_calls([
            mock.call(obj, 'FLAV', 'update'),
            mock.call(obj, 'FLAV', 'update'),
        ])
        self.assertEqual(0, cc.server_resize.call_count)

    def test__update_flavor_resize_failed(self):
        obj = mock.Mock(physical_id='NOVA_ID')
        cc = mock.Mock()
        cc.server_resize.side_effect = [
            exc.InternalError(code=500, message='Resize failed')]
        profile = server.ServerProfile('t', self.spec)
        profile._computeclient = cc
        new_spec = copy.deepcopy(self.spec)
        new_spec['properties']['flavor'] = 'new_flavor'
        new_profile = server.ServerProfile('t1', new_spec)
        x_flavors = [mock.Mock(id='123'), mock.Mock(id='456')]
        mock_validate = self.patchobject(profile, '_validate_flavor',
                                         side_effect=x_flavors)

        ex = self.assertRaises(exc.EResourceUpdate,
                               profile._update_flavor,
                               obj, new_profile)

        mock_validate.assert_has_calls([
            mock.call(obj, 'FLAV', 'update'),
            mock.call(obj, 'new_flavor', 'update'),
        ])
        cc.server_resize.assert_called_once_with('NOVA_ID', '456')
        cc.server_resize_revert.assert_called_once_with('NOVA_ID')
        cc.wait_for_server.assert_called_once_with('NOVA_ID', 'ACTIVE')
        self.assertEqual("Failed in updating server 'NOVA_ID': Resize "
                         "failed.", six.text_type(ex))

    def test__update_flavor_first_wait_for_server_failed(self):
        obj = mock.Mock(physical_id='NOVA_ID')
        cc = mock.Mock()
        cc.wait_for_server.side_effect = [
            exc.InternalError(code=500, message='TIMEOUT'),
            None
        ]

        profile = server.ServerProfile('t', self.spec)
        profile._computeclient = cc
        new_spec = copy.deepcopy(self.spec)
        new_spec['properties']['flavor'] = 'new_flavor'
        new_profile = server.ServerProfile('t1', new_spec)
        x_flavors = [mock.Mock(id='123'), mock.Mock(id='456')]
        mock_validate = self.patchobject(profile, '_validate_flavor',
                                         side_effect=x_flavors)
        # do it
        ex = self.assertRaises(exc.EResourceUpdate,
                               profile._update_flavor,
                               obj, new_profile)

        # assertions
        mock_validate.assert_has_calls([
            mock.call(obj, 'FLAV', 'update'),
            mock.call(obj, 'new_flavor', 'update'),
        ])
        cc.server_resize.assert_called_once_with('NOVA_ID', '456')
        cc.wait_for_server.has_calls([
            mock.call('NOVA_ID', 'VERIFY_RESIZE'),
            mock.call('NOVA_ID', 'ACTIVE')])
        cc.server_resize_revert.assert_called_once_with('NOVA_ID')
        self.assertEqual("Failed in updating server 'NOVA_ID': "
                         "TIMEOUT.", six.text_type(ex))

    def test__update_flavor_resize_failed_revert_failed(self):
        obj = mock.Mock(physical_id='NOVA_ID')
        cc = mock.Mock()
        err_resize = exc.InternalError(code=500, message='Resize')
        cc.server_resize.side_effect = err_resize
        err_revert = exc.InternalError(code=500, message='Revert')
        cc.server_resize_revert.side_effect = err_revert
        profile = server.ServerProfile('t', self.spec)
        profile._computeclient = cc
        new_spec = copy.deepcopy(self.spec)
        new_spec['properties']['flavor'] = 'new_flavor'
        new_profile = server.ServerProfile('t1', new_spec)
        x_flavors = [mock.Mock(id='123'), mock.Mock(id='456')]
        mock_validate = self.patchobject(profile, '_validate_flavor',
                                         side_effect=x_flavors)

        # do it
        ex = self.assertRaises(exc.EResourceUpdate,
                               profile._update_flavor,
                               obj, new_profile)

        # assertions
        mock_validate.assert_has_calls([
            mock.call(obj, 'FLAV', 'update'),
            mock.call(obj, 'new_flavor', 'update'),
        ])
        cc.server_resize.assert_called_once_with('NOVA_ID', '456')
        cc.server_resize_revert.assert_called_once_with('NOVA_ID')
        # the wait_for_server wasn't called
        self.assertEqual(0, cc.wait_for_server.call_count)
        self.assertEqual("Failed in updating server 'NOVA_ID': "
                         "Revert.", six.text_type(ex))

    def test__update_flavor_confirm_failed(self):
        obj = mock.Mock(physical_id='NOVA_ID')
        cc = mock.Mock()
        err_confirm = exc.InternalError(code=500, message='Confirm')
        cc.server_resize_confirm.side_effect = err_confirm
        profile = server.ServerProfile('t', self.spec)
        profile._computeclient = cc
        new_spec = copy.deepcopy(self.spec)
        new_spec['properties']['flavor'] = 'new_flavor'
        new_profile = server.ServerProfile('t1', new_spec)
        x_flavors = [mock.Mock(id='123'), mock.Mock(id='456')]
        mock_validate = self.patchobject(profile, '_validate_flavor',
                                         side_effect=x_flavors)

        # do it
        ex = self.assertRaises(exc.EResourceUpdate,
                               profile._update_flavor,
                               obj, new_profile)

        # assertions
        mock_validate.assert_has_calls([
            mock.call(obj, 'FLAV', 'update'),
            mock.call(obj, 'new_flavor', 'update'),
        ])
        cc.server_resize.assert_called_once_with('NOVA_ID', '456')
        cc.server_resize_confirm.assert_called_once_with('NOVA_ID')
        cc.wait_for_server.assert_called_once_with('NOVA_ID', 'VERIFY_RESIZE')
        self.assertEqual("Failed in updating server 'NOVA_ID': Confirm.",
                         six.text_type(ex))

    def test__update_flavor_wait_confirm_failed(self):
        obj = mock.Mock(physical_id='NOVA_ID')
        cc = mock.Mock()
        err_wait = exc.InternalError(code=500, message='Wait')
        cc.wait_for_server.side_effect = [None, err_wait]
        profile = server.ServerProfile('t', self.spec)
        profile._computeclient = cc
        new_spec = copy.deepcopy(self.spec)
        new_spec['properties']['flavor'] = 'new_flavor'
        new_profile = server.ServerProfile('t1', new_spec)
        x_flavors = [mock.Mock(id='123'), mock.Mock(id='456')]
        mock_validate = self.patchobject(profile, '_validate_flavor',
                                         side_effect=x_flavors)

        # do it
        ex = self.assertRaises(exc.InternalError,
                               profile._update_flavor,
                               obj, new_profile)

        # assertions
        mock_validate.assert_has_calls([
            mock.call(obj, 'FLAV', 'update'),
            mock.call(obj, 'new_flavor', 'update'),
        ])
        cc.server_resize.assert_called_once_with('NOVA_ID', '456')
        cc.server_resize_confirm.assert_called_once_with('NOVA_ID')
        cc.wait_for_server.assert_has_calls([
            mock.call('NOVA_ID', 'VERIFY_RESIZE'),
            mock.call('NOVA_ID', 'ACTIVE')
        ])
        self.assertEqual("Failed in updating server 'NOVA_ID': Wait.",
                         six.text_type(ex))

    def test__update_image(self):
        profile = server.ServerProfile('t', self.spec)
        x_image = {'id': '123'}
        x_server = mock.Mock(image=x_image)
        cc = mock.Mock()
        cc.server_get.return_value = x_server
        profile._computeclient = cc
        x_new_image = mock.Mock(id='456')
        x_images = [x_new_image]
        mock_check = self.patchobject(profile, '_validate_image',
                                      side_effect=x_images)
        obj = mock.Mock(physical_id='NOVA_ID')
        new_spec = copy.deepcopy(self.spec)
        new_spec['properties']['image'] = 'new_image'
        new_profile = server.ServerProfile('t1', new_spec)

        profile._update_image(obj, new_profile, 'new_name', 'new_pass')

        mock_check.assert_has_calls([
            mock.call(obj, 'new_image', reason='update'),
        ])
        cc.server_rebuild.assert_called_once_with(
            'NOVA_ID', '456', 'new_name', 'new_pass')
        cc.wait_for_server.assert_called_once_with('NOVA_ID', 'ACTIVE')

    def test__update_image_new_image_is_none(self):
        profile = server.ServerProfile('t', self.spec)
        obj = mock.Mock(physical_id='NOVA_ID')
        new_spec = copy.deepcopy(self.spec)
        del new_spec['properties']['image']
        new_profile = server.ServerProfile('t1', new_spec)

        ex = self.assertRaises(exc.EResourceUpdate,
                               profile._update_image,
                               obj, new_profile, 'new_name', '')

        msg = ("Failed in updating server 'NOVA_ID': Updating Nova server"
               " with image set to None is not supported by Nova.")
        self.assertEqual(msg, six.text_type(ex))

    def test__update_image_new_image_invalid(self):
        # NOTE: The image invalid could be caused by a non-existent image or
        # a compute driver failure
        profile = server.ServerProfile('t', self.spec)
        # _validate_image will always throw EResourceUpdate if driver fails
        err = exc.EResourceUpdate(type='server', id='NOVA_ID', message='BAD')
        mock_check = self.patchobject(profile, '_validate_image',
                                      side_effect=err)
        obj = mock.Mock(physical_id='NOVA_ID')
        new_spec = copy.deepcopy(self.spec)
        new_spec['properties']['image'] = 'new_image'
        new_profile = server.ServerProfile('t1', new_spec)

        ex = self.assertRaises(exc.EResourceUpdate,
                               profile._update_image,
                               obj, new_profile, 'new_name', 'new_pass')

        msg = ("Failed in updating server 'NOVA_ID': BAD.")
        self.assertEqual(msg, six.text_type(ex))
        mock_check.assert_called_once_with(obj, 'new_image', reason='update')

    def test__update_image_old_image_invalid(self):
        # NOTE: The image invalid could be caused by a non-existent image or
        # a compute driver failure
        profile = server.ServerProfile('t', self.spec)
        x_image = {'id': '123'}
        x_server = mock.Mock(image=x_image)
        cc = mock.Mock()
        cc.server_get.return_value = x_server
        profile._computeclient = cc
        # _validate_image will always throw EResourceUpdate if driver fails
        results = [
            exc.EResourceUpdate(type='server', id='NOVA_ID', message='BAD')
        ]
        mock_check = self.patchobject(profile, '_validate_image',
                                      side_effect=results)
        obj = mock.Mock(physical_id='NOVA_ID')
        new_spec = copy.deepcopy(self.spec)
        new_spec['properties']['image'] = 'new_image'
        new_profile = server.ServerProfile('t1', new_spec)

        ex = self.assertRaises(exc.EResourceUpdate,
                               profile._update_image,
                               obj, new_profile, 'new_name', 'new_pass')

        msg = ("Failed in updating server 'NOVA_ID': BAD.")
        self.assertEqual(msg, six.text_type(ex))
        mock_check.assert_has_calls([
            mock.call(obj, 'new_image', reason='update'),
        ])

    def test__update_image_old_image_is_none_but_succeeded(self):
        old_spec = copy.deepcopy(self.spec)
        del old_spec['properties']['image']
        profile = server.ServerProfile('t', old_spec)
        cc = mock.Mock()
        profile._computeclient = cc
        x_server = mock.Mock(image={'id': '123'})
        cc.server_get.return_value = x_server
        # this is the new one
        x_image = mock.Mock(id='456')
        mock_check = self.patchobject(profile, '_validate_image',
                                      return_value=x_image)
        obj = mock.Mock(physical_id='NOVA_ID')
        new_spec = copy.deepcopy(self.spec)
        new_spec['properties']['image'] = 'new_image'
        new_profile = server.ServerProfile('t1', new_spec)

        res = profile._update_image(obj, new_profile, 'new_name', 'new_pass')

        self.assertTrue(res)
        mock_check.assert_called_once_with(obj, 'new_image', reason='update')
        cc.server_get.assert_called_once_with('NOVA_ID')
        cc.server_rebuild.assert_called_once_with(
            'NOVA_ID', '456', 'new_name', 'new_pass')
        cc.wait_for_server.assert_called_once_with('NOVA_ID', 'ACTIVE')

    def test__update_image_old_image_is_none_but_failed(self):
        old_spec = copy.deepcopy(self.spec)
        del old_spec['properties']['image']
        profile = server.ServerProfile('t', old_spec)
        cc = mock.Mock()
        profile._computeclient = cc
        # this is about the new one
        x_image = mock.Mock(id='456')
        mock_check = self.patchobject(profile, '_validate_image',
                                      return_value=x_image)
        cc.server_get.side_effect = exc.InternalError(message='DRIVER')
        obj = mock.Mock(physical_id='NOVA_ID')
        new_spec = copy.deepcopy(self.spec)
        new_spec['properties']['image'] = 'new_image'
        new_profile = server.ServerProfile('t1', new_spec)

        ex = self.assertRaises(exc.EResourceUpdate,
                               profile._update_image,
                               obj, new_profile, 'new_name', 'new_pass')

        self.assertEqual("Failed in updating server 'NOVA_ID': DRIVER.",
                         six.text_type(ex))
        mock_check.assert_called_once_with(obj, 'new_image', reason='update')
        cc.server_get.assert_called_once_with('NOVA_ID')

    def test__update_image_updating_to_same_image(self):
        profile = server.ServerProfile('t', self.spec)
        x_image = {'id': '123'}
        x_server = mock.Mock(image=x_image)
        cc = mock.Mock()
        cc.server_get.return_value = x_server
        profile._computeclient = cc
        x_new_image = mock.Mock(id='123')
        x_images = [x_new_image]
        mock_check = self.patchobject(profile, '_validate_image',
                                      side_effect=x_images)
        obj = mock.Mock(physical_id='NOVA_ID')
        new_spec = copy.deepcopy(self.spec)
        new_spec['properties']['image'] = 'new_image'
        new_profile = server.ServerProfile('t1', new_spec)

        res = profile._update_image(obj, new_profile, 'new_name', 'new_pass')

        self.assertFalse(res)
        mock_check.assert_has_calls([
            mock.call(obj, 'new_image', reason='update'),
        ])
        self.assertEqual(0, cc.server_rebuild.call_count)
        self.assertEqual(0, cc.wait_for_server.call_count)

    def test__update_image_failed_rebuilding(self):
        profile = server.ServerProfile('t', self.spec)
        x_image = {'id': '123'}
        x_server = mock.Mock(image=x_image)
        cc = mock.Mock()
        cc.server_get.return_value = x_server
        cc.server_rebuild.side_effect = exc.InternalError(message='FAILED')
        profile._computeclient = cc
        x_new_image = mock.Mock(id='456')
        x_images = [x_new_image]
        mock_check = self.patchobject(profile, '_validate_image',
                                      side_effect=x_images)
        obj = mock.Mock(physical_id='NOVA_ID')
        new_spec = copy.deepcopy(self.spec)
        new_spec['properties']['image'] = 'new_image'
        new_profile = server.ServerProfile('t1', new_spec)

        ex = self.assertRaises(exc.EResourceUpdate,
                               profile._update_image,
                               obj, new_profile, 'new_name', 'new_pass')

        self.assertEqual("Failed in updating server 'NOVA_ID': FAILED.",
                         six.text_type(ex))
        mock_check.assert_has_calls([
            mock.call(obj, 'new_image', reason='update'),
        ])
        cc.server_rebuild.assert_called_once_with(
            'NOVA_ID', '456', 'new_name', 'new_pass')
        self.assertEqual(0, cc.wait_for_server.call_count)

    def test__update_image_failed_waiting(self):
        profile = server.ServerProfile('t', self.spec)
        x_image = {'id': '123'}
        x_server = mock.Mock(image=x_image)
        cc = mock.Mock()
        cc.server_get.return_value = x_server
        cc.wait_for_server.side_effect = exc.InternalError(message='TIMEOUT')
        profile._computeclient = cc
        x_new_image = mock.Mock(id='456')
        x_images = [x_new_image]
        mock_check = self.patchobject(profile, '_validate_image',
                                      side_effect=x_images)
        obj = mock.Mock(physical_id='NOVA_ID')
        new_spec = copy.deepcopy(self.spec)
        new_spec['properties']['image'] = 'new_image'
        new_profile = server.ServerProfile('t1', new_spec)

        ex = self.assertRaises(exc.EResourceUpdate,
                               profile._update_image,
                               obj, new_profile, 'new_name', 'new_pass')

        self.assertEqual("Failed in updating server 'NOVA_ID': TIMEOUT.",
                         six.text_type(ex))
        mock_check.assert_has_calls([
            mock.call(obj, 'new_image', reason='update'),
        ])
        cc.server_rebuild.assert_called_once_with(
            'NOVA_ID', '456', 'new_name', 'new_pass')
        cc.wait_for_server.assert_called_once_with('NOVA_ID', 'ACTIVE')

    def test__create_interfaces(self):
        cc = mock.Mock()
        server_obj = mock.Mock()
        cc.server_get.return_value = server_obj
        profile = server.ServerProfile('t', self.spec)
        profile._computeclient = cc
        validation_results = [
            {'network': 'net1_id', 'fixed_ip': 'ip2'},
            {'network': 'net2_id'},
            {'port': 'port4'}
        ]
        mock_validate = self.patchobject(profile, '_validate_network',
                                         side_effect=validation_results)
        ports_results = [
            (mock.Mock(
                id='port1_id', network_id='net1_id',
                fixed_ips=[{'ip_address': 'ip2'}], security_group_ids=[]),
             None),
            (mock.Mock(
                id='port2_id', network_id='net2_id',
                fixed_ips=[{'ip_address': 'ip3'}], security_group_ids=[]),
             None),
            (mock.Mock(
                id='port4_id', network_id='net3_id',
                fixed_ips=[{'ip_address': 'ip4'}], security_group_ids=[]),
             None)
        ]
        mock_get_port = self.patchobject(profile, '_get_port',
                                         side_effect=ports_results)
        networks = [
            {'network': 'net1', 'port': None, 'fixed_ip': 'ip2'},
            {'network': 'net2', 'port': None, 'fixed_ip': None},
            {'network': None, 'port': 'port4', 'fixed_ip': None}
        ]
        obj = mock.Mock(physical_id='NOVA_ID', data={})

        res = profile._update_network_add_port(obj, networks)

        self.assertIsNone(res)
        cc.server_get.assert_called_with('NOVA_ID')
        validation_calls = [
            mock.call(obj,
                      {'network': 'net1', 'port': None, 'fixed_ip': 'ip2'},
                      'update'),
            mock.call(obj,
                      {'network': 'net2', 'port': None, 'fixed_ip': None},
                      'update'),
            mock.call(obj,
                      {'network': None, 'port': 'port4', 'fixed_ip': None},
                      'update')
        ]
        mock_validate.assert_has_calls(validation_calls)
        mock_get_port.assert_called_with(obj, {'port': 'port4'})
        create_calls = [
            mock.call(server_obj, port='port1_id'),
            mock.call(server_obj, port='port2_id'),
            mock.call(server_obj, port='port4_id'),
        ]
        cc.server_interface_create.assert_has_calls(create_calls)

    def test__create_interfaces_failed_getting_server(self):
        cc = mock.Mock()
        cc.server_get.side_effect = exc.InternalError(message='Not valid')
        profile = server.ServerProfile('t', self.spec)
        profile._computeclient = cc
        self.patchobject(profile, '_create_ports_from_properties')

        obj = mock.Mock(physical_id='NOVA_ID')
        networks = [{'foo': 'bar'}]  # not used

        ex = self.assertRaises(exc.EResourceUpdate,
                               profile._update_network_add_port,
                               obj, networks)

        self.assertEqual("Failed in updating server 'NOVA_ID': Not valid.",
                         six.text_type(ex))
        cc.server_get.assert_called_once_with('NOVA_ID')
        self.assertEqual(0, profile._create_ports_from_properties.call_count)

    def test__create_interfaces_failed_validation(self):
        cc = mock.Mock()
        server_obj = mock.Mock()
        cc.server_get.return_value = server_obj
        profile = server.ServerProfile('t', self.spec)
        profile._computeclient = cc
        err = exc.EResourceUpdate(type='server', id='NOVA_ID',
                                  message='Driver error')
        mock_validate = self.patchobject(profile, '_validate_network',
                                         side_effect=err)
        networks = [{'network': 'net1', 'port': None, 'fixed_ip': 'ip2'}]
        obj = mock.Mock(physical_id='NOVA_ID')

        ex = self.assertRaises(exc.EResourceUpdate,
                               profile._update_network_add_port,
                               obj, networks)

        self.assertEqual("Failed in updating server 'NOVA_ID': Driver error.",
                         six.text_type(ex))
        cc.server_get.assert_called_once_with('NOVA_ID')
        mock_validate.assert_called_once_with(obj, networks[0], 'update')
        self.assertEqual(0, cc.server_interface_create.call_count)

    def test__delete_interfaces(self):
        cc = mock.Mock()
        nc = mock.Mock()
        net1 = mock.Mock(id='net1')
        nc.network_get.return_value = net1
        nc.port_find.return_value = mock.Mock(id='port3', status='DOWN')
        profile = server.ServerProfile('t', self.spec)
        profile._computeclient = cc
        profile._networkclient = nc
        obj = mock.Mock(physical_id='NOVA_ID', data={'internal_ports': [
            {'id': 'port1', 'network_id': 'net1', 'remove': True,
             'fixed_ips': [{'ip_address': 'ip1'}]},
            {'id': 'port2', 'network_id': 'net1', 'remove': True,
             'fixed_ips': [{'ip_address': 'ip-random2'}]},
            {'id': 'port3', 'network_id': 'net1', 'remove': True,
             'fixed_ips': [{'ip_address': 'ip3'}]}]})
        networks = [
            {'network': 'net1', 'port': None, 'fixed_ip': 'ip1'},
            {'network': 'net1', 'port': None, 'fixed_ip': None},
            {'network': None, 'port': 'port3', 'fixed_ip': None}
        ]

        res = profile._update_network_remove_port(obj, networks)

        self.assertIsNone(res)

        nc.network_get.assert_has_calls([
            mock.call('net1'), mock.call('net1')
        ])
        cc.server_interface_delete.assert_has_calls([
            mock.call('port1', 'NOVA_ID'),
            mock.call('port2', 'NOVA_ID'),
            mock.call('port3', 'NOVA_ID'),
        ])
        nc.port_delete.assert_has_calls([
            mock.call('port1', ignore_missing=True),
            mock.call('port2', ignore_missing=True),
            mock.call('port3', ignore_missing=True),
        ])

    def test__delete_interfaces_failed_delete(self):
        cc = mock.Mock()
        profile = server.ServerProfile('t', self.spec)
        profile._computeclient = cc
        profile._networkclient = mock.Mock()
        candidate_ports = [
            [{'id': 'port1', 'network_id': 'net1',
             'fixed_ips': [{'ip_address': 'ip1'}]}],
        ]
        self.patchobject(profile, '_find_port_by_net_spec',
                         side_effect=candidate_ports)
        err = exc.InternalError(message='BANG')
        cc.server_interface_delete.side_effect = err
        internal_ports = [
            {'id': 'port1', 'remove': True}
        ]
        obj = mock.Mock(physical_id='NOVA_ID',
                        data={'internal_ports': internal_ports})
        networks = [
            {'network': 'net1', 'port': None, 'fixed_ip': 'ip1'},
        ]

        ex = self.assertRaises(exc.EResourceUpdate,
                               profile._update_network_remove_port,
                               obj, networks)

        self.assertEqual("Failed in updating server 'NOVA_ID': BANG.",
                         six.text_type(ex))
        cc.server_interface_delete.assert_called_once_with('port1', 'NOVA_ID')

    @mock.patch.object(server.ServerProfile, '_update_network_remove_port')
    @mock.patch.object(server.ServerProfile, '_update_network_add_port')
    def test__update_network(self, mock_create, mock_delete):
        obj = mock.Mock(physical_id='FAKE_ID')

        old_spec = copy.deepcopy(self.spec)
        old_spec['properties']['networks'] = [
            {'network': 'net1', 'fixed_ip': 'ip1'},
            {'network': 'net1'},
            {'port': 'port3'},
        ]
        profile = server.ServerProfile('t', old_spec)
        new_spec = copy.deepcopy(self.spec)
        new_spec['properties']['networks'] = [
            {'network': 'net1', 'fixed_ip': 'ip2'},
            {'network': 'net2'},
            {'port': 'port4'},
        ]
        new_profile = server.ServerProfile('t1', new_spec)

        res = profile._update_network(obj, new_profile)

        self.assertIsNone(res)

        networks_create = [
            {'floating_network': None, 'network': 'net1', 'fixed_ip': 'ip2',
             'floating_ip': None, 'port': None, 'security_groups': None},
            {'floating_network': None, 'network': 'net2', 'fixed_ip': None,
             'floating_ip': None, 'port': None, 'security_groups': None},
            {'floating_network': None, 'network': None, 'fixed_ip': None,
             'floating_ip': None, 'port': 'port4', 'security_groups': None}
        ]
        mock_create.assert_called_once_with(obj, networks_create)
        networks_delete = [
            {'floating_network': None, 'network': 'net1', 'fixed_ip': 'ip1',
             'floating_ip': None, 'port': None, 'security_groups': None},
            {'floating_network': None, 'network': 'net1', 'fixed_ip': None,
             'floating_ip': None, 'port': None, 'security_groups': None},
            {'floating_network': None, 'network': None, 'fixed_ip': None,
             'floating_ip': None, 'port': 'port3', 'security_groups': None}
        ]
        mock_delete.assert_called_once_with(obj, networks_delete)

    @mock.patch.object(server.ServerProfile, '_update_password')
    @mock.patch.object(server.ServerProfile, '_check_password')
    @mock.patch.object(server.ServerProfile, '_update_name')
    @mock.patch.object(server.ServerProfile, '_check_server_name')
    @mock.patch.object(server.ServerProfile, '_update_flavor')
    @mock.patch.object(server.ServerProfile, '_update_metadata')
    @mock.patch.object(server.ServerProfile, '_update_image')
    @mock.patch.object(server.ServerProfile, '_update_network')
    def test_do_update_name_succeeded(self, mock_update_network,
                                      mock_update_image, mock_update_metadata,
                                      mock_update_flavor, mock_check_name,
                                      mock_update_name, mock_check_password,
                                      mock_update_password):
        mock_check_name.return_value = True, 'NEW_NAME'
        mock_check_password.return_value = True, 'NEW_PASSWORD'
        mock_update_image.return_value = False
        obj = mock.Mock(physical_id='FAKE_ID')

        profile = server.ServerProfile('t', self.spec)
        profile._computeclient = mock.Mock()
        new_profile = server.ServerProfile('t', self.spec)

        res = profile.do_update(obj, new_profile)

        self.assertTrue(res)
        mock_check_name.assert_called_once_with(obj, new_profile)
        mock_update_metadata.assert_called_once_with(obj, new_profile)
        mock_update_image.assert_called_once_with(
            obj, new_profile, 'NEW_NAME', 'NEW_PASSWORD')
        mock_update_name.assert_called_once_with(obj, 'NEW_NAME')
        mock_update_password.assert_called_once_with(obj, 'NEW_PASSWORD')
        mock_update_flavor.assert_called_once_with(obj, new_profile)
        mock_update_network.assert_called_once_with(obj, new_profile)

    @mock.patch.object(server.ServerProfile, '_update_password')
    @mock.patch.object(server.ServerProfile, '_check_password')
    @mock.patch.object(server.ServerProfile, '_update_name')
    @mock.patch.object(server.ServerProfile, '_check_server_name')
    @mock.patch.object(server.ServerProfile, '_update_flavor')
    @mock.patch.object(server.ServerProfile, '_update_metadata')
    @mock.patch.object(server.ServerProfile, '_update_image')
    @mock.patch.object(server.ServerProfile, '_update_network')
    def test_do_update_name_no_change(self, mock_update_network,
                                      mock_update_image, mock_update_metadata,
                                      mock_update_flavor, mock_check_name,
                                      mock_update_name, mock_check_password,
                                      mock_update_password):
        mock_check_name.return_value = False, 'NEW_NAME'
        mock_check_password.return_value = False, 'OLD_PASS'
        obj = mock.Mock(physical_id='NOVA_ID')

        profile = server.ServerProfile('t', self.spec)
        profile._computeclient = mock.Mock()
        new_profile = server.ServerProfile('t', self.spec)

        res = profile.do_update(obj, new_profile)

        self.assertTrue(res)
        mock_check_name.assert_called_once_with(obj, new_profile)
        mock_check_password.assert_called_once_with(obj, new_profile)
        mock_update_image.assert_called_once_with(
            obj, new_profile, 'NEW_NAME', 'OLD_PASS')
        self.assertEqual(0, mock_update_name.call_count)
        self.assertEqual(0, mock_update_password.call_count)
        mock_update_flavor.assert_called_once_with(obj, new_profile)
        mock_update_network.assert_called_once_with(obj, new_profile)
        mock_update_metadata.assert_called_once_with(obj, new_profile)

    @mock.patch.object(server.ServerProfile, '_update_password')
    @mock.patch.object(server.ServerProfile, '_check_password')
    @mock.patch.object(server.ServerProfile, '_update_name')
    @mock.patch.object(server.ServerProfile, '_check_server_name')
    @mock.patch.object(server.ServerProfile, '_update_flavor')
    @mock.patch.object(server.ServerProfile, '_update_metadata')
    @mock.patch.object(server.ServerProfile, '_update_image')
    @mock.patch.object(server.ServerProfile, '_update_network')
    def test_do_update_name_failed(self, mock_update_network,
                                   mock_update_image, mock_update_metadata,
                                   mock_update_flavor, mock_check_name,
                                   mock_update_name, mock_check_password,
                                   mock_update_password):
        mock_check_name.return_value = True, 'NEW_NAME'
        mock_check_password.return_value = False, 'OLD_PASS'
        mock_update_image.return_value = False
        err = exc.EResourceUpdate(type='server', id='NOVA_ID', message='BANG')
        mock_update_name.side_effect = err
        obj = mock.Mock(physical_id='NOVA_ID')

        profile = server.ServerProfile('t', self.spec)
        profile._computeclient = mock.Mock()
        new_profile = server.ServerProfile('t', self.spec)

        ex = self.assertRaises(exc.EResourceUpdate,
                               profile.do_update,
                               obj, new_profile)

        self.assertEqual("Failed in updating server 'NOVA_ID': BANG.",
                         six.text_type(ex))
        mock_check_name.assert_called_once_with(obj, new_profile)
        mock_check_password.assert_called_once_with(obj, new_profile)
        mock_update_image.assert_called_once_with(
            obj, new_profile, 'NEW_NAME', 'OLD_PASS')
        mock_update_name.assert_called_once_with(obj, 'NEW_NAME')
        self.assertEqual(0, mock_update_password.call_count)
        self.assertEqual(0, mock_update_flavor.call_count)
        self.assertEqual(0, mock_update_metadata.call_count)

    @mock.patch.object(server.ServerProfile, '_update_password')
    @mock.patch.object(server.ServerProfile, '_update_name')
    @mock.patch.object(server.ServerProfile, '_check_password')
    @mock.patch.object(server.ServerProfile, '_check_server_name')
    @mock.patch.object(server.ServerProfile, '_update_flavor')
    @mock.patch.object(server.ServerProfile, '_update_image')
    def test_do_update_image_succeeded(self, mock_update_image,
                                       mock_update_flavor, mock_check_name,
                                       mock_check_password, mock_update_name,
                                       mock_update_password):
        mock_check_name.return_value = False, 'OLD_NAME'
        mock_check_password.return_value = False, 'OLD_PASS'
        mock_update_image.return_value = True
        obj = mock.Mock()
        obj.physical_id = 'FAKE_ID'

        profile = server.ServerProfile('t', self.spec)
        profile._computeclient = mock.Mock()
        new_spec = copy.deepcopy(self.spec)
        new_spec['properties']['image'] = 'FAKE_IMAGE_NEW'
        new_profile = server.ServerProfile('t', new_spec)

        res = profile.do_update(obj, new_profile)

        self.assertTrue(res)
        mock_update_image.assert_called_with(
            obj, new_profile, 'OLD_NAME', 'OLD_PASS')
        self.assertEqual(0, mock_update_name.call_count)
        self.assertEqual(0, mock_update_password.call_count)

    @mock.patch.object(server.ServerProfile, '_update_flavor')
    @mock.patch.object(server.ServerProfile, '_update_name')
    @mock.patch.object(server.ServerProfile, '_update_metadata')
    @mock.patch.object(server.ServerProfile, '_update_image')
    @mock.patch.object(server.ServerProfile, '_check_password')
    @mock.patch.object(server.ServerProfile, '_check_server_name')
    def test_do_update_image_failed(self, mock_check_name, mock_check_password,
                                    mock_update_image, mock_update_meta,
                                    mock_update_name, mock_update_flavor):
        mock_check_name.return_value = False, 'OLD_NAME'
        mock_check_password.return_value = False, 'OLD_PASS'
        # _update_image always throw EResourceUpdate
        ex = exc.EResourceUpdate(type='server', id='NOVA_ID',
                                 message='Image Not Found')
        mock_update_image.side_effect = ex
        obj = mock.Mock(physical_id='NOVA_ID')

        profile = server.ServerProfile('t', self.spec)
        profile._computeclient = mock.Mock()
        # don't need to invent a new spec
        new_spec = copy.deepcopy(self.spec)
        new_profile = server.ServerProfile('t', new_spec)

        ex = self.assertRaises(exc.EResourceUpdate,
                               profile.do_update,
                               obj, new_profile)

        mock_update_image.assert_called_with(
            obj, new_profile, 'OLD_NAME', 'OLD_PASS')
        self.assertEqual("Failed in updating server 'NOVA_ID': "
                         "Image Not Found.", six.text_type(ex))

    @mock.patch.object(server.ServerProfile, '_update_flavor')
    def test_do_update_update_flavor_succeeded(self, mock_update_flavor):
        obj = mock.Mock(physical_id='FAKE_ID')
        profile = server.ServerProfile('t', self.spec)
        x_image = {'id': '123'}
        x_server = mock.Mock(image=x_image)
        cc = mock.Mock()
        cc.server_get.return_value = x_server
        profile._computeclient = cc
        new_spec = copy.deepcopy(self.spec)
        new_spec['properties']['flavor'] = 'FAKE_FLAVOR_NEW'
        new_profile = server.ServerProfile('t', new_spec)

        res = profile.do_update(obj, new_profile)
        self.assertTrue(res)
        mock_update_flavor.assert_called_with(obj, new_profile)

    @mock.patch.object(server.ServerProfile, '_update_flavor')
    def test_do_update_update_flavor_failed(self, mock_update_flavor):
        ex = exc.EResourceUpdate(type='server', id='NOVA_ID',
                                 message='Flavor Not Found')
        mock_update_flavor.side_effect = ex
        obj = mock.Mock(physical_id='NOVA_ID')
        profile = server.ServerProfile('t', self.spec)
        x_image = {'id': '123'}
        x_server = mock.Mock(image=x_image)
        cc = mock.Mock()
        cc.server_get.return_value = x_server
        profile._computeclient = cc
        new_spec = copy.deepcopy(self.spec)
        new_spec['properties']['flavor'] = 'FAKE_FLAVOR_NEW'
        new_profile = server.ServerProfile('t', new_spec)

        ex = self.assertRaises(exc.EResourceUpdate,
                               profile.do_update,
                               obj, new_profile)

        mock_update_flavor.assert_called_with(obj, new_profile)
        self.assertEqual("Failed in updating server 'NOVA_ID': "
                         "Flavor Not Found.",
                         six.text_type(ex))

    @mock.patch.object(server.ServerProfile, '_update_flavor')
    @mock.patch.object(server.ServerProfile, '_update_network')
    def test_do_update_update_network_succeeded(
            self, mock_update_network, mock_update_flavor):
        mock_update_network.return_value = True
        profile = server.ServerProfile('t', self.spec)
        x_image = {'id': '123'}
        x_server = mock.Mock(image=x_image)
        cc = mock.Mock()
        cc.server_get.return_value = x_server
        profile._computeclient = cc

        obj = mock.Mock(physical_id='NOVA_ID')

        new_spec = copy.deepcopy(self.spec)
        new_spec['properties']['networks'] = [
            {'network': 'new_net', 'port': 'new_port', 'fixed_ip': 'new-ip'}
        ]
        new_profile = server.ServerProfile('t', new_spec)

        res = profile.do_update(obj, new_profile)
        self.assertTrue(res)
        mock_update_network.assert_called_with(obj, new_profile)

    @mock.patch.object(server.ServerProfile, '_update_password')
    @mock.patch.object(server.ServerProfile, '_check_password')
    @mock.patch.object(server.ServerProfile, '_update_name')
    @mock.patch.object(server.ServerProfile, '_check_server_name')
    @mock.patch.object(server.ServerProfile, '_update_flavor')
    @mock.patch.object(server.ServerProfile, '_update_metadata')
    @mock.patch.object(server.ServerProfile, '_update_image')
    @mock.patch.object(server.ServerProfile, '_update_network')
    def test_do_update_update_network_failed(
            self, mock_update_network, mock_update_image, mock_update_metadata,
            mock_update_flavor, mock_check_name, mock_update_name,
            mock_check_password, mock_update_password):

        mock_check_name.return_value = True, 'NEW_NAME'
        mock_check_password.return_value = True, 'NEW_PASSWORD'
        mock_update_image.return_value = True
        err = exc.EResourceUpdate(type='server', id='NOVA_ID', message='BOOM')
        mock_update_network.side_effect = err

        profile = server.ServerProfile('t', self.spec)
        profile._computeclient = mock.Mock()
        new_network = {
            'port': 'FAKE_PORT_NEW',
            'fixed_ip': 'FAKE_IP_NEW',
            'network': 'FAKE_NET_NEW',
        }
        new_spec = copy.deepcopy(self.spec)
        new_spec['properties']['networks'] = [new_network]
        new_profile = server.ServerProfile('t', new_spec)
        obj = mock.Mock(physical_id='NOVA_ID')

        ex = self.assertRaises(exc.EResourceUpdate,
                               profile.do_update,
                               obj, new_profile)

        self.assertEqual("Failed in updating server 'NOVA_ID': BOOM.",
                         six.text_type(ex))
        mock_check_name.assert_called_once_with(obj, new_profile)
        mock_check_password.assert_called_once_with(obj, new_profile)
        mock_update_image.assert_called_once_with(
            obj, new_profile, 'NEW_NAME', 'NEW_PASSWORD')
        self.assertEqual(0, mock_update_name.call_count)
        self.assertEqual(0, mock_update_password.call_count)
        mock_update_flavor.assert_called_once_with(obj, new_profile)
        mock_update_network.assert_called_with(obj, new_profile)
        self.assertEqual(0, mock_update_metadata.call_count)

    def test_do_update_without_profile(self):
        profile = server.ServerProfile('t', self.spec)
        obj = mock.Mock()
        obj.physical_id = 'FAKE_ID'
        new_profile = None
        res = profile.do_update(obj, new_profile)
        self.assertFalse(res)

    def test_do_update_no_physical_id(self):
        profile = server.ServerProfile('t', self.spec)
        profile._computeclient = mock.Mock()
        node_obj = mock.Mock(physical_id=None)
        new_profile = mock.Mock()

        # Test path where server doesn't exist
        res = profile.do_update(node_obj, new_profile)

        self.assertFalse(res)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import mock

from senlin.common import consts
from senlin.common import exception as exc
from senlin.engine.actions import base as ab
from senlin.engine.actions import cluster_action as ca
from senlin.engine import cluster as cm
from senlin.engine import dispatcher
from senlin.engine import senlin_lock
from senlin.policies import base as pb
from senlin.tests.unit.common import base
from senlin.tests.unit.common import utils

CLUSTER_ID = 'e1cfd82b-dc95-46ad-86e8-37864d7be1cd'
OBJID = '571fffb8-f41c-4cbc-945c-cb2937d76f19'
ACTION_ID = '4c2cead2-fd74-418a-9d12-bd2d9bd7a812'


@mock.patch.object(cm.Cluster, 'load')
class ClusterActionTest(base.SenlinTestCase):

    def setUp(self):
        super(ClusterActionTest, self).setUp()
        self.ctx = utils.dummy_context()

    @mock.patch.object(ab.Action, 'policy_check')
    def test__execute(self, mock_check, mock_load):
        cluster = mock.Mock()
        cluster.id = 'FAKE_CLUSTER'
        mock_load.return_value = cluster
        action = ca.ClusterAction(cluster.id, 'CLUSTER_FLY', self.ctx)
        action.do_fly = mock.Mock(return_value=(action.RES_OK, 'Good!'))
        action.data = {
            'status': pb.CHECK_OK,
            'reason': 'Policy checking passed'
        }

        res_code, res_msg = action._execute()

        self.assertEqual(action.RES_OK, res_code)
        self.assertEqual('Good!', res_msg)
        mock_check.assert_has_calls([
            mock.call('FAKE_CLUSTER', 'BEFORE'),
            mock.call('FAKE_CLUSTER', 'AFTER')])

    @mock.patch.object(ab.Action, 'policy_check')
    def test_execute_failed_policy_check(self, mock_check, mock_load):
        cluster = mock.Mock()
        cluster.id = 'FAKE_CLUSTER'
        mock_load.return_value = cluster
        action = ca.ClusterAction(cluster.id, 'CLUSTER_FLY', self.ctx)
        action.do_fly = mock.Mock(return_value=(action.RES_OK, 'Good!'))
        action.data = {
            'status': pb.CHECK_ERROR,
            'reason': 'Something is wrong.'
        }

        res_code, res_msg = action._execute()

        self.assertEqual(action.RES_ERROR, res_code)
        self.assertEqual('Policy check failure: Something is wrong.', res_msg)
        mock_check.assert_called_once_with('FAKE_CLUSTER', 'BEFORE')

    @mock.patch.object(ab.Action, 'policy_check')
    def test_execute_unsupported_action(self, mock_check, mock_load):
        cluster = mock.Mock()
        cluster.id = 'FAKE_CLUSTER'
        mock_load.return_value = cluster
        action = ca.ClusterAction(cluster.id, 'CLUSTER_DANCE', self.ctx)
        action.data = {
            'status': pb.CHECK_OK,
            'reason': 'All is going well.'
        }

        res_code, res_msg = action._execute()

        self.assertEqual(action.RES_ERROR, res_code)
        self.assertEqual('Unsupported action: CLUSTER_DANCE.', res_msg)
        mock_check.assert_called_once_with('FAKE_CLUSTER', 'BEFORE')

    def test_execute_post_check_failed(self, mock_load):
        def fake_check(cluster_id, target):
            if target == 'BEFORE':
                action.data = {
                    'status': pb.CHECK_OK,
                    'reason': 'Policy checking passed.'
                }
            else:
                action.data = {
                    'status': pb.CHECK_ERROR,
                    'reason': 'Policy checking failed.'
                }

        cluster = mock.Mock()
        cluster.id = 'FAKE_CLUSTER'
        mock_load.return_value = cluster
        action = ca.ClusterAction(cluster.id, 'CLUSTER_FLY', self.ctx)
        action.do_fly = mock.Mock(return_value=(action.RES_OK, 'Cool!'))
        mock_check = self.patchobject(action, 'policy_check',
                                      side_effect=fake_check)

        res_code, res_msg = action._execute()

        self.assertEqual(action.RES_ERROR, res_code)
        self.assertEqual('Policy check failure: Policy checking failed.',
                         res_msg)
        mock_check.assert_has_calls([
            mock.call('FAKE_CLUSTER', 'BEFORE'),
            mock.call('FAKE_CLUSTER', 'AFTER')])

    @mock.patch.object(senlin_lock, 'cluster_lock_acquire')
    @mock.patch.object(senlin_lock, 'cluster_lock_release')
    def test_execute(self, mock_release, mock_acquire, mock_load):
        cluster = mock.Mock()
        cluster.id = 'FAKE_CLUSTER'
        mock_load.return_value = cluster
        action = ca.ClusterAction(cluster.id, 'CLUSTER_FLY', self.ctx)
        action.id = 'ACTION_ID'
        self.patchobject(action, '_execute',
                         return_value=(action.RES_OK, 'success'))
        mock_acquire.return_value = action

        res_code, res_msg = action.execute()

        self.assertEqual(action.RES_OK, res_code)
        self.assertEqual('success', res_msg)
        mock_load.assert_has_calls(
            [mock.call(action.context, 'FAKE_CLUSTER'),
             mock.call(action.context, 'FAKE_CLUSTER')])
        mock_acquire.assert_called_once_with(
            self.ctx, 'FAKE_CLUSTER', 'ACTION_ID', None,
            senlin_lock.CLUSTER_SCOPE, False)
        mock_release.assert_called_once_with(
            'FAKE_CLUSTER', 'ACTION_ID', senlin_lock.CLUSTER_SCOPE)

    @mock.patch.object(senlin_lock, 'cluster_lock_acquire')
    def test_execute_failed_locking(self, mock_acquire, mock_load):
        cluster = mock.Mock()
        cluster.id = 'CLUSTER_ID'
        mock_load.return_value = cluster
        action = ca.ClusterAction(cluster.id, 'CLUSTER_DELETE', self.ctx)
        mock_acquire.return_value = None

        res_code, res_msg = action.execute()

        self.assertEqual(action.RES_RETRY, res_code)
        self.assertEqual('Failed in locking cluster.', res_msg)
        mock_load.assert_called_once_with(action.context, cluster.id)

    @mock.patch.object(senlin_lock, 'cluster_lock_acquire')
    @mock.patch.object(senlin_lock, 'cluster_lock_release')
    def test_execute_failed_execute(self, mock_release, mock_acquire,
                                    mock_load):
        cluster = mock.Mock()
        cluster.id = 'CLUSTER_ID'
        mock_load.return_value = cluster
        action = ca.ClusterAction(cluster.id, 'CLUSTER_DELETE', self.ctx)
        action.id = 'ACTION_ID'
        mock_acquire.return_value = action
        self.patchobject(action, '_execute',
                         return_value=(action.RES_ERROR, 'Failed execution.'))

        res_code, res_msg = action.execute()

        self.assertEqual(action.RES_ERROR, res_code)
        self.assertEqual('Failed execution.', res_msg)
        mock_load.assert_has_calls(
            [mock.call(action.context, cluster.id),
             mock.call(action.context, cluster.id)])
        mock_acquire.assert_called_once_with(
            self.ctx, 'CLUSTER_ID', 'ACTION_ID', None,
            senlin_lock.CLUSTER_SCOPE, True)
        mock_release.assert_called_once_with(
            'CLUSTER_ID', 'ACTION_ID', senlin_lock.CLUSTER_SCOPE)

    def test_cancel(self, mock_load):
        action = ca.ClusterAction('ID', 'CLUSTER_DELETE', self.ctx)
        res = action.cancel()
        self.assertEqual(action.RES_OK, res)


class CompleteLifecycleProcTest(base.SenlinTestCase):

    def setUp(self):
        super(CompleteLifecycleProcTest, self).setUp()

        self.ctx = utils.dummy_context()

    @mock.patch.object(dispatcher, 'start_action')
    @mock.patch.object(ab.Action, 'load')
    def test_complete_lifecycle_proc_successful(self, mock_load,
                                                mock_dispatcher_start):
        action = ab.Action(OBJID, 'OBJECT_ACTION', self.ctx)
        mock_obj = mock.Mock()
        action.entity = mock_obj
        mock_get_status = self.patchobject(action, 'get_status')
        mock_get_status.return_value = \
            consts.ACTION_WAITING_LIFECYCLE_COMPLETION
        mock_set_status = self.patchobject(action, 'set_status')
        mock_load.return_value = action

        res = ca.CompleteLifecycleProc(self.ctx, 'ACTION_ID')

        self.assertTrue(res)
        mock_load.assert_called_once_with(self.ctx, action_id='ACTION_ID',
                                          project_safe=False)
        mock_get_status.assert_called_once_with()
        mock_set_status.assert_called_once_with(action.RES_LIFECYCLE_COMPLETE)
        mock_dispatcher_start.assert_called_once_with()

    @mock.patch.object(ab.Action, 'load')
    def test_complete_lifecycle_proc_failed_action_not_found(self, mock_load):
        mock_load.return_value = None

        self.assertRaises(exc.ResourceNotFound,
                          ca.CompleteLifecycleProc,
                          self.ctx, 'ACTION')

    @mock.patch.object(dispatcher, 'start_action')
    @mock.patch.object(ab.Action, 'load')
    def test_complete_lifecycle_proc_warning(self, mock_load,
                                             mock_dispatcher_start):
        action = ab.Action(OBJID, 'OBJECT_ACTION', self.ctx)
        mock_obj = mock.Mock()
        action.entity = mock_obj
        mock_get_status = self.patchobject(action, 'get_status')
        mock_get_status.return_value = consts.ACTION_SUCCEEDED
        mock_set_status = self.patchobject(action, 'set_status')
        mock_load.return_value = action

        res = ca.CompleteLifecycleProc(self.ctx, 'ACTION_ID')

        self.assertFalse(res)
        mock_load.assert_called_once_with(self.ctx, action_id='ACTION_ID',
                                          project_safe=False)
        mock_get_status.assert_called_once_with()
        mock_set_status.assert_not_called()
        mock_dispatcher_start.assert_not_called()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import eventlet
import six
import time

from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import timeutils

from senlin.common import consts
from senlin.common import context as req_context
from senlin.common import exception
from senlin.common import utils
from senlin.engine import dispatcher
from senlin.engine import event as EVENT
from senlin.objects import action as ao
from senlin.objects import cluster_policy as cpo
from senlin.objects import dependency as dobj
from senlin.policies import base as policy_mod

wallclock = time.time
LOG = logging.getLogger(__name__)


class Action(object):
    """An action can be performed on a cluster or a node of a cluster."""

    RETURNS = (
        RES_OK, RES_ERROR, RES_RETRY, RES_CANCEL, RES_TIMEOUT,
        RES_LIFECYCLE_COMPLETE, RES_LIFECYCLE_HOOK_TIMEOUT,
    ) = (
        'OK', 'ERROR', 'RETRY', 'CANCEL', 'TIMEOUT', 'LIFECYCLE_COMPLETE',
        'LIFECYCLE_HOOK_TIMEOUT'
    )

    # Action status definitions:
    #  INIT:      Not ready to be executed because fields are being modified,
    #             or dependency with other actions are being analyzed.
    #  READY:     Initialized and ready to be executed by a worker.
    #  RUNNING:   Being executed by a worker thread.
    #  SUCCEEDED: Completed with success.
    #  FAILED:    Completed with failure.
    #  CANCELLED: Action cancelled because worker thread was cancelled.
    STATUSES = (
        INIT, WAITING, READY, RUNNING, SUSPENDED,
        SUCCEEDED, FAILED, CANCELLED, WAITING_LIFECYCLE_COMPLETION
    ) = (
        'INIT', 'WAITING', 'READY', 'RUNNING', 'SUSPENDED',
        'SUCCEEDED', 'FAILED', 'CANCELLED', 'WAITING_LIFECYCLE_COMPLETION'
    )

    # Signal commands
    COMMANDS = (
        SIG_CANCEL, SIG_SUSPEND, SIG_RESUME,
    ) = (
        'CANCEL', 'SUSPEND', 'RESUME',
    )

    def __new__(cls, target, action, ctx, **kwargs):
        if (cls != Action):
            return super(Action, cls).__new__(cls)

        target_type = action.split('_')[0]
        if target_type == 'CLUSTER':
            from senlin.engine.actions import cluster_action
            ActionClass = cluster_action.ClusterAction
        elif target_type == 'NODE':
            from senlin.engine.actions import node_action
            ActionClass = node_action.NodeAction
        else:
            from senlin.engine.actions import custom_action
            ActionClass = custom_action.CustomAction

        return super(Action, cls).__new__(ActionClass)

    def __init__(self, target, action, ctx, **kwargs):
        # context will be persisted into database so that any worker thread
        # can pick the action up and execute it on behalf of the initiator

        self.id = kwargs.get('id', None)
        self.name = kwargs.get('name', '')

        self.context = ctx
        self.user = ctx.user_id
        self.project = ctx.project_id
        self.domain = ctx.domain_id

        self.action = action
        self.target = target

        # Why this action is fired, it can be a UUID of another action
        self.cause = kwargs.get('cause', '')

        # Owner can be an UUID format ID for the worker that is currently
        # working on the action.  It also serves as a lock.
        self.owner = kwargs.get('owner', None)

        # An action may need to be executed repeatitively, interval is the
        # time in seconds between two consecutive execution.
        # A value of -1 indicates that this action is only to be executed once
        self.interval = kwargs.get('interval', -1)

        # Start time can be an absolute time or a time relative to another
        # action. E.g.
        #   - '2014-12-18 08:41:39.908569'
        #   - 'AFTER: 57292917-af90-4c45-9457-34777d939d4d'
        #   - 'WHEN: 0265f93b-b1d7-421f-b5ad-cb83de2f559d'
        self.start_time = kwargs.get('start_time', None)
        self.end_time = kwargs.get('end_time', None)

        # Timeout is a placeholder in case some actions may linger too long
        self.timeout = kwargs.get('timeout', cfg.CONF.default_action_timeout)

        # Return code, useful when action is not automatically deleted
        # after execution
        self.status = kwargs.get('status', self.INIT)
        self.status_reason = kwargs.get('status_reason', '')

        # All parameters are passed in using keyword arguments which is
        # a dictionary stored as JSON in DB
        self.inputs = kwargs.get('inputs', {})
        self.outputs = kwargs.get('outputs', {})

        self.created_at = kwargs.get('created_at', None)
        self.updated_at = kwargs.get('updated_at', None)

        self.data = kwargs.get('data', {})

    def store(self, ctx):
        """Store the action record into database table.

        :param ctx: An instance of the request context.
        :return: The ID of the stored object.
        """

        timestamp = timeutils.utcnow(True)

        values = {
            'name': self.name,
            'context': self.context.to_dict(),
            'target': self.target,
            'action': self.action,
            'cause': self.cause,
            'owner': self.owner,
            'interval': self.interval,
            'start_time': self.start_time,
            'end_time': self.end_time,
            'timeout': self.timeout,
            'status': self.status,
            'status_reason': self.status_reason,
            'inputs': self.inputs,
            'outputs': self.outputs,
            'created_at': self.created_at,
            'updated_at': self.updated_at,
            'data': self.data,
            'user': self.user,
            'project': self.project,
            'domain': self.domain,
        }

        if self.id:
            self.updated_at = timestamp
            values['updated_at'] = timestamp
            ao.Action.update(ctx, self.id, values)
        else:
            self.created_at = timestamp
            values['created_at'] = timestamp
            action = ao.Action.create(ctx, values)
            self.id = action.id

        return self.id

    @classmethod
    def _from_object(cls, obj):
        """Construct an action from database object.

        :param obj: a DB action object that contains all fields.
        :return: An `Action` object deserialized from the DB action object.
        """
        ctx = req_context.RequestContext.from_dict(obj.context)
        kwargs = {
            'id': obj.id,
            'name': obj.name,
            'cause': obj.cause,
            'owner': obj.owner,
            'interval': obj.interval,
            'start_time': obj.start_time,
            'end_time': obj.end_time,
            'timeout': obj.timeout,
            'status': obj.status,
            'status_reason': obj.status_reason,
            'inputs': obj.inputs or {},
            'outputs': obj.outputs or {},
            'created_at': obj.created_at,
            'updated_at': obj.updated_at,
            'data': obj.data,
        }

        target_type = obj.action.split('_')[0]
        if target_type == 'CLUSTER':
            from senlin.engine.actions import cluster_action
            ActionClass = cluster_action.ClusterAction
        elif target_type == 'NODE':
            from senlin.engine.actions import node_action
            ActionClass = node_action.NodeAction
        else:
            from senlin.engine.actions import custom_action
            ActionClass = custom_action.CustomAction

        return ActionClass(obj.target, obj.action, ctx, **kwargs)

    @classmethod
    def load(cls, ctx, action_id=None, db_action=None, project_safe=True):
        """Retrieve an action from database.

        :param ctx: Instance of request context.
        :param action_id: An UUID for the action to deserialize.
        :param db_action: An action object for the action to deserialize.
        :return: A `Action` object instance.
        """
        if db_action is None:
            db_action = ao.Action.get(ctx, action_id,
                                      project_safe=project_safe)
            if db_action is None:
                raise exception.ResourceNotFound(type='action', id=action_id)

        return cls._from_object(db_action)

    @classmethod
    def create(cls, ctx, target, action, **kwargs):
        """Create an action object.

        :param ctx: The requesting context.
        :param target: The ID of the target cluster/node.
        :param action: Name of the action.
        :param dict kwargs: Other keyword arguments for the action.
        :return: ID of the action created.
        """
        params = {
            'user_id': ctx.user_id,
            'project_id': ctx.project_id,
            'domain_id': ctx.domain_id,
            'is_admin': ctx.is_admin,
            'request_id': ctx.request_id,
            'trusts': ctx.trusts,
        }
        c = req_context.RequestContext.from_dict(params)
        obj = cls(target, action, c, **kwargs)
        return obj.store(ctx)

    @classmethod
    def delete(cls, ctx, action_id):
        """Delete an action from database.

        :param ctx: An instance of the request context.
        :param action_id: The UUID of the target action to be deleted.
        :return: Nothing.
        """
        ao.Action.delete(ctx, action_id)

    def signal(self, cmd):
        """Send a signal to the action.

        :param cmd: One of the command word defined in self.COMMANDS.
        :returns: None
        """
        if cmd not in self.COMMANDS:
            return

        if cmd == self.SIG_CANCEL:
            expected = (self.INIT, self.WAITING, self.READY, self.RUNNING)
        elif cmd == self.SIG_SUSPEND:
            expected = (self.RUNNING)
        else:  # SIG_RESUME
            expected = (self.SUSPENDED)

        if self.status not in expected:
            LOG.error("Action (%(id)s) is in status (%(actual)s) while "
                      "expected status must be one of (%(expected)s).",
                      dict(id=self.id[:8], expected=expected,
                           actual=self.status))
            return

        ao.Action.signal(self.context, self.id, cmd)

    def execute(self, **kwargs):
        """Execute the action.

        In theory, the action encapsulates all information needed for
        execution.  'kwargs' may specify additional parameters.
        :param kwargs: additional parameters that may override the default
                       properties stored in the action record.
        """
        raise NotImplementedError

    def set_status(self, result, reason=None):
        """Set action status based on return value from execute."""

        timestamp = wallclock()

        if result == self.RES_OK:
            status = self.SUCCEEDED
            ao.Action.mark_succeeded(self.context, self.id, timestamp)

        elif result == self.RES_ERROR:
            status = self.FAILED
            ao.Action.mark_failed(self.context, self.id, timestamp,
                                  reason or 'ERROR')

        elif result == self.RES_TIMEOUT:
            status = self.FAILED
            ao.Action.mark_failed(self.context, self.id, timestamp,
                                  reason or 'TIMEOUT')

        elif result == self.RES_CANCEL:
            status = self.CANCELLED
            ao.Action.mark_cancelled(self.context, self.id, timestamp)

        elif result == self.RES_LIFECYCLE_COMPLETE:
            status = self.SUCCEEDED
            ao.Action.mark_ready(self.context, self.id, timestamp)

        else:  # result == self.RES_RETRY:
            retries = self.data.get('retries', 0)
            # Action failed at the moment, but can be retried
            # retries time is configurable
            if retries < cfg.CONF.lock_retry_times:
                status = self.READY
                retries += 1

                self.data.update({'retries': retries})
                ao.Action.abandon(self.context, self.id, {'data': self.data})
                # sleep for a while
                eventlet.sleep(cfg.CONF.lock_retry_interval)
                dispatcher.start_action(self.id)
            else:
                status = self.RES_ERROR
                if not reason:
                    reason = ('Exceeded maximum number of retries (%d)'
                              '') % cfg.CONF.lock_retry_times
                ao.Action.mark_failed(self.context, self.id, timestamp, reason)

        if status == self.SUCCEEDED:
            EVENT.info(self, consts.PHASE_END, reason or 'SUCCEEDED')
        elif status == self.READY:
            EVENT.warning(self, consts.PHASE_ERROR, reason or 'RETRY')
        else:
            EVENT.error(self, consts.PHASE_ERROR, reason or 'ERROR')

        self.status = status
        self.status_reason = reason

    def get_status(self):
        timestamp = wallclock()
        status = ao.Action.check_status(self.context, self.id, timestamp)
        self.status = status
        return status

    def is_timeout(self, timeout=None):
        if timeout is None:
            timeout = self.timeout
        time_elapse = wallclock() - self.start_time
        return time_elapse > timeout

    def _check_signal(self):
        # Check timeout first, if true, return timeout message
        if self.timeout is not None and self.is_timeout():
            EVENT.debug(self, consts.PHASE_ERROR, 'TIMEOUT')
            return self.RES_TIMEOUT

        result = ao.Action.signal_query(self.context, self.id)
        return result

    def is_cancelled(self):
        return self._check_signal() == self.SIG_CANCEL

    def is_suspended(self):
        return self._check_signal() == self.SIG_SUSPEND

    def is_resumed(self):
        return self._check_signal() == self.SIG_RESUME

    def _check_result(self, name):
        """Check policy status and generate event.

        :param name: Name of policy checked
        :return: True if the policy checking can be continued, or False if the
                 policy checking should be aborted.
        """
        reason = self.data['reason']
        if self.data['status'] == policy_mod.CHECK_OK:
            return True

        self.data['reason'] = ("Failed policy '%(name)s': %(reason)s"
                               ) % {'name': name, 'reason': reason}
        return False

    def policy_check(self, cluster_id, target):
        """Check all policies attached to cluster and give result.

        :param cluster_id: The ID of the cluster to which the policy is
            attached.
        :param target: A tuple of ('when', action_name)
        :return: A dictionary that contains the check result.
        """

        if target not in ['BEFORE', 'AFTER']:
            return

        bindings = cpo.ClusterPolicy.get_all(self.context, cluster_id,
                                             sort='priority',
                                             filters={'enabled': True})
        # default values
        self.data['status'] = policy_mod.CHECK_OK
        self.data['reason'] = 'Completed policy checking.'

        for pb in bindings:
            policy = policy_mod.Policy.load(self.context, pb.policy_id)
            # We record the last operation time for all policies bound to the
            # cluster, no matter that policy is only interested in the
            # "BEFORE" or "AFTER" or both.
            if target == 'AFTER':
                ts = timeutils.utcnow(True)
                pb.last_op = ts
                cpo.ClusterPolicy.update(self.context, pb.cluster_id,
                                         pb.policy_id, {'last_op': ts})

            if not policy.need_check(target, self):
                continue

            if target == 'BEFORE':
                method = getattr(policy, 'pre_op', None)
            else:  # target == 'AFTER'
                method = getattr(policy, 'post_op', None)

            if getattr(policy, 'cooldown', None):
                if pb.cooldown_inprogress(policy.cooldown):
                    self.data['status'] = policy_mod.CHECK_ERROR
                    self.data['reason'] = ('Policy %s cooldown is still '
                                           'in progress.') % policy.id
                    return

            if method is not None:
                method(cluster_id, self)

            res = self._check_result(policy.name)
            if res is False:
                return
        return

    def to_dict(self):
        if self.id:
            dep_on = dobj.Dependency.get_depended(self.context, self.id)
            dep_by = dobj.Dependency.get_dependents(self.context, self.id)
        else:
            dep_on = []
            dep_by = []
        action_dict = {
            'id': self.id,
            'name': self.name,
            'action': self.action,
            'target': self.target,
            'cause': self.cause,
            'owner': self.owner,
            'interval': self.interval,
            'start_time': self.start_time,
            'end_time': self.end_time,
            'timeout': self.timeout,
            'status': self.status,
            'status_reason': self.status_reason,
            'inputs': self.inputs,
            'outputs': self.outputs,
            'depends_on': dep_on,
            'depended_by': dep_by,
            'created_at': utils.isotime(self.created_at),
            'updated_at': utils.isotime(self.updated_at),
            'data': self.data,
            'user': self.user,
            'project': self.project,
        }
        return action_dict


def ActionProc(ctx, action_id):
    """Action process."""

    # Step 1: materialize the action object
    action = Action.load(ctx, action_id=action_id, project_safe=False)
    if action is None:
        LOG.error('Action "%s" could not be found.', action_id)
        return False

    EVENT.info(action, consts.PHASE_START, action_id[:8])

    reason = 'Action completed'
    success = True
    try:
        # Step 2: execute the action
        result, reason = action.execute()
        if result == action.RES_RETRY:
            success = False
    except Exception as ex:
        # We catch exception here to make sure the following logics are
        # executed.
        result = action.RES_ERROR
        reason = six.text_type(ex)
        LOG.exception('Unexpected exception occurred during action '
                      '%(action)s (%(id)s) execution: %(reason)s',
                      {'action': action.action, 'id': action.id,
                       'reason': reason})
        success = False
    finally:
        # NOTE: locks on action is eventually released here by status update
        action.set_status(result, reason)

    return success



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import mock
import six

from senlin.common import consts
from senlin.common import exception as exc
from senlin.common import scaleutils as su
from senlin.engine import cluster as cm
from senlin.policies import base as pb
from senlin.policies import region_placement as rp
from senlin.tests.unit.common import base
from senlin.tests.unit.common import utils


class TestRegionPlacementPolicy(base.SenlinTestCase):

    def setUp(self):
        super(TestRegionPlacementPolicy, self).setUp()
        self.context = utils.dummy_context()
        self.spec = {
            'type': 'senlin.policy.region_placement',
            'version': '1.0',
            'properties': {
                'regions': [
                    {'name': 'R1', 'weight': 100, 'cap': 50},
                    {'name': 'R2', 'weight': 50, 'cap': 50},
                    {'name': 'R3', 'weight': 30, 'cap': -1},
                    {'name': 'R4', 'weight': 20, 'cap': -1}
                ]
            }
        }

    def test_policy_init(self):
        policy = rp.RegionPlacementPolicy('test-policy', self.spec)

        self.assertIsNone(policy.id)
        self.assertIsNone(policy. _keystoneclient)
        self.assertEqual('test-policy', policy.name)
        self.assertEqual('senlin.policy.region_placement-1.0', policy.type)
        expected = {
            'R1': {
                'weight': 100,
                'cap': 50
            },
            'R2': {
                'weight': 50,
                'cap': 50,
            },
            'R3': {
                'weight': 30,
                'cap': -1,
            },
            'R4': {
                'weight': 20,
                'cap': -1,
            }
        }
        self.assertEqual(expected, policy.regions)

    @mock.patch.object(pb.Policy, 'validate')
    def test_validate_okay(self, mock_base_validate):
        policy = rp.RegionPlacementPolicy('test-policy', self.spec)
        kc = mock.Mock()
        kc.validate_regions.return_value = ['R1', 'R2', 'R3', 'R4']
        policy._keystoneclient = kc
        ctx = mock.Mock(user='U1', project='P1')

        res = policy.validate(ctx, True)

        self.assertTrue(res)
        mock_base_validate.assert_called_once_with(ctx, True)
        kc.validate_regions.assert_called_once_with(['R1', 'R2', 'R3', 'R4'])

    @mock.patch.object(pb.Policy, 'validate')
    def test_validate_no_validate_props(self, mock_base_validate):
        policy = rp.RegionPlacementPolicy('test-policy', self.spec)
        ctx = mock.Mock(user='U1', project='P1')

        res = policy.validate(ctx, False)

        self.assertTrue(res)
        mock_base_validate.assert_called_once_with(ctx, False)

    @mock.patch.object(pb.Policy, 'validate')
    def test_validate_region_not_found(self, mock_base_validate):
        policy = rp.RegionPlacementPolicy('test-policy', self.spec)
        kc = mock.Mock()
        kc.validate_regions.return_value = ['R2', 'R4']
        policy._keystoneclient = kc
        ctx = mock.Mock(user='U1', project='P1')

        ex = self.assertRaises(exc.InvalidSpec,
                               policy.validate,
                               ctx, True)

        mock_base_validate.assert_called_once_with(ctx, True)
        kc.validate_regions.assert_called_once_with(['R1', 'R2', 'R3', 'R4'])
        self.assertEqual("The specified regions '['R1', 'R3']' could not "
                         "be found.", six.text_type(ex))

    def test__create_plan(self):
        policy = rp.RegionPlacementPolicy('p1', self.spec)
        regions = policy.regions

        current = {'R1': 2, 'R2': 2, 'R3': 2, 'R4': 1}
        result = policy._create_plan(current, regions, 5, True)
        expected = {'R1': 4, 'R2': 1}
        self.assertEqual(expected, result)

        current = {'R1': 2, 'R2': 2, 'R3': 0, 'R4': 1}
        plan = policy._create_plan(current, regions, 5, True)
        answer = {'R1': 3, 'R2': 1, 'R3': 1}
        self.assertEqual(answer, plan)

        current = {'R1': 2, 'R2': 2, 'R3': 0, 'R4': 1}
        plan = policy._create_plan(current, regions, 3, False)
        answer = {'R2': 2, 'R4': 1}
        self.assertEqual(answer, plan)

        current = {'R1': 4, 'R2': 2, 'R3': 1, 'R4': 1}
        plan = policy._create_plan(current, regions, 3, False)
        answer = {'R2': 1, 'R3': 1, 'R4': 1}
        self.assertEqual(answer, plan)

    def test__get_count_node_create_no_region(self):
        x_profile = mock.Mock(CONTEXT='context', properties={'context': {}})
        x_node = mock.Mock(rt={'profile': x_profile})
        action = mock.Mock(action=consts.NODE_CREATE, entity=x_node)

        policy = rp.RegionPlacementPolicy('p1', self.spec)

        res = policy._get_count('FOO', action)
        self.assertEqual(1, res)

    def test__get_count_node_create_region_specified(self):
        x_profile = mock.Mock(CONTEXT='context',
                              properties={'context': {'region_name': 'foo'}})
        x_node = mock.Mock(rt={'profile': x_profile})
        action = mock.Mock(action=consts.NODE_CREATE, entity=x_node)

        policy = rp.RegionPlacementPolicy('p1', self.spec)

        res = policy._get_count('FOO', action)
        self.assertEqual(0, res)

    def test__get_count_resize_deletion(self):
        action = mock.Mock(action=consts.CLUSTER_RESIZE,
                           data={'deletion': {'count': 3}})

        policy = rp.RegionPlacementPolicy('p1', self.spec)

        res = policy._get_count('FOO', action)
        self.assertEqual(-3, res)

    def test__get_count_resize_creation(self):
        action = mock.Mock(action=consts.CLUSTER_RESIZE,
                           data={'creation': {'count': 3}})
        policy = rp.RegionPlacementPolicy('p1', self.spec)

        res = policy._get_count('FOO', action)

        self.assertEqual(3, res)

    @mock.patch.object(su, 'parse_resize_params')
    def test__get_count_resize_parse_error(self, mock_parse):
        x_cluster = mock.Mock()
        x_cluster.nodes = [mock.Mock(), mock.Mock()]
        action = mock.Mock(action=consts.CLUSTER_RESIZE, data={})
        action.entity = x_cluster
        mock_parse.return_value = (pb.CHECK_ERROR, 'Something wrong.')
        policy = rp.RegionPlacementPolicy('p1', self.spec)

        res = policy._get_count('FOO', action)

        self.assertEqual(0, res)
        self.assertEqual(pb.CHECK_ERROR, action.data['status'])
        mock_parse.assert_called_once_with(action, x_cluster, 2)
        self.assertEqual('Something wrong.', action.data['reason'])

    @mock.patch.object(su, 'parse_resize_params')
    def test__get_count_resize_parse_creation(self, mock_parse):
        def fake_parse(action, cluster, current):
            action.data = {'creation': {'count': 3}}
            return pb.CHECK_OK, ''

        x_cluster = mock.Mock()
        x_cluster.nodes = []
        action = mock.Mock(action=consts.CLUSTER_RESIZE, data={})
        action.entity = x_cluster

        mock_parse.side_effect = fake_parse
        policy = rp.RegionPlacementPolicy('p1', self.spec)

        res = policy._get_count('FOO', action)

        self.assertEqual(3, res)
        mock_parse.assert_called_once_with(action, x_cluster, 0)

    @mock.patch.object(su, 'parse_resize_params')
    def test__get_count_resize_parse_deletion(self, mock_parse):
        def fake_parse(action, cluster, current):
            action.data = {'deletion': {'count': 3}}
            return pb.CHECK_OK, ''

        x_cluster = mock.Mock()
        x_cluster.nodes = [mock.Mock(), mock.Mock(), mock.Mock()]
        action = mock.Mock(action=consts.CLUSTER_RESIZE, data={})
        action.entity = x_cluster

        mock_parse.side_effect = fake_parse
        policy = rp.RegionPlacementPolicy('p1', self.spec)

        res = policy._get_count('FOO', action)

        self.assertEqual(-3, res)
        mock_parse.assert_called_once_with(action, x_cluster, 3)

    def test__get_count_scale_in_with_data(self):
        action = mock.Mock(action=consts.CLUSTER_SCALE_IN,
                           data={'deletion': {'count': 3}})
        policy = rp.RegionPlacementPolicy('p1', self.spec)

        res = policy._get_count('FOO', action)
        self.assertEqual(-3, res)

    def test__get_count_scale_in_with_no_data(self):
        action = mock.Mock(action=consts.CLUSTER_SCALE_IN,
                           data={'deletion': {'num': 3}})
        policy = rp.RegionPlacementPolicy('p1', self.spec)

        res = policy._get_count('FOO', action)
        self.assertEqual(-1, res)

    def test__get_count_scale_in_with_inputs(self):
        action = mock.Mock(action=consts.CLUSTER_SCALE_IN, data={},
                           inputs={'count': 3})
        policy = rp.RegionPlacementPolicy('p1', self.spec)

        res = policy._get_count('FOO', action)
        self.assertEqual(-3, res)

    def test__get_count_scale_in_with_incorrect_inputs(self):
        action = mock.Mock(action=consts.CLUSTER_SCALE_IN, data={},
                           inputs={'num': 3})
        policy = rp.RegionPlacementPolicy('p1', self.spec)

        res = policy._get_count('FOO', action)
        self.assertEqual(-1, res)

    def test__get_count_scale_out_with_data(self):
        action = mock.Mock(action=consts.CLUSTER_SCALE_OUT,
                           data={'creation': {'count': 3}})
        policy = rp.RegionPlacementPolicy('p1', self.spec)

        res = policy._get_count('FOO', action)
        self.assertEqual(3, res)

    def test__get_count_scale_out_with_no_data(self):
        action = mock.Mock(action=consts.CLUSTER_SCALE_OUT,
                           data={'creation': {'num': 3}})
        policy = rp.RegionPlacementPolicy('p1', self.spec)

        res = policy._get_count('FOO', action)
        self.assertEqual(1, res)

    def test__get_count_scale_out_with_inputs(self):
        action = mock.Mock(action=consts.CLUSTER_SCALE_OUT, data={},
                           inputs={'count': 3})
        policy = rp.RegionPlacementPolicy('p1', self.spec)

        res = policy._get_count('FOO', action)
        self.assertEqual(3, res)

    def test__get_count_scale_out_with_incorrect_inputs(self):
        action = mock.Mock(action=consts.CLUSTER_SCALE_OUT, data={},
                           inputs={'num': 3})
        policy = rp.RegionPlacementPolicy('p1', self.spec)

        res = policy._get_count('FOO', action)
        self.assertEqual(1, res)

    @mock.patch.object(cm.Cluster, 'load')
    def test_pre_op(self, mock_load):
        # test pre_op method whether returns the correct action.data
        policy = rp.RegionPlacementPolicy('p1', self.spec)
        regions = policy.regions

        kc = mock.Mock()
        kc.validate_regions.return_value = regions.keys()
        policy._keystoneclient = kc

        plan = {'R1': 1, 'R3': 2}
        self.patchobject(policy, '_create_plan', return_value=plan)

        action = mock.Mock()
        action.context = self.context
        action.action = 'CLUSTER_SCALE_OUT'
        action.inputs = {}
        action.data = {
            'creation': {
                'count': 3,
            }
        }

        cluster = mock.Mock()
        current_dist = {'R1': 0, 'R2': 0, 'R3': 0, 'R4': 0}
        cluster.get_region_distribution.return_value = current_dist
        mock_load.return_value = cluster

        res = policy.pre_op('FAKE_CLUSTER', action)

        self.assertIsNone(res)

        self.assertEqual(3, action.data['creation']['count'])
        dist = action.data['creation']['regions']
        self.assertEqual(2, len(dist))
        self.assertEqual(1, dist['R1'])
        self.assertEqual(2, dist['R3'])

        mock_load.assert_called_once_with(action.context, 'FAKE_CLUSTER')
        kc.validate_regions.assert_called_once_with(regions.keys())
        cluster.get_region_distribution.assert_called_once_with(regions.keys())
        policy._create_plan.assert_called_once_with(
            current_dist, regions, 3, True)

    @mock.patch.object(cm.Cluster, 'load')
    def test_pre_op_count_from_inputs(self, mock_load):
        # test pre_op method whether returns the correct action.data
        policy = rp.RegionPlacementPolicy('p1', self.spec)
        regions = policy.regions

        kc = mock.Mock()
        kc.validate_regions.return_value = regions.keys()
        policy._keystoneclient = kc

        cluster = mock.Mock()
        current_dist = {'R1': 0, 'R2': 0, 'R3': 0, 'R4': 0}
        cluster.get_region_distribution.return_value = current_dist
        mock_load.return_value = cluster

        plan = {'R1': 1, 'R3': 2}
        self.patchobject(policy, '_create_plan', return_value=plan)

        action = mock.Mock()
        action.context = self.context
        action.action = 'CLUSTER_SCALE_OUT'
        action.inputs = {'count': 3}
        action.data = {}

        res = policy.pre_op('FAKE_CLUSTER', action)

        self.assertIsNone(res)
        self.assertEqual(3, action.data['creation']['count'])
        dist = action.data['creation']['regions']
        self.assertEqual(2, len(dist))
        self.assertEqual(1, dist['R1'])
        self.assertEqual(2, dist['R3'])

    @mock.patch.object(cm.Cluster, 'load')
    def test_pre_op_no_regions(self, mock_load):
        # test pre_op method whether returns the correct action.data
        policy = rp.RegionPlacementPolicy('p1', self.spec)
        kc = mock.Mock()
        kc.validate_regions.return_value = []
        policy._keystoneclient = kc

        action = mock.Mock()
        action.action = 'CLUSTER_SCALE_OUT'
        action.context = self.context
        action.data = {'creation': {'count': 3}}

        cluster = mock.Mock()
        mock_load.return_value = cluster

        res = policy.pre_op('FAKE_CLUSTER', action)

        self.assertIsNone(res)
        self.assertEqual('ERROR', action.data['status'])
        self.assertEqual('No region is found usable.', action.data['reason'])

    @mock.patch.object(cm.Cluster, 'load')
    def test_pre_op_no_feasible_plan(self, mock_load):
        # test pre_op method whether returns the correct action.data
        policy = rp.RegionPlacementPolicy('p1', self.spec)
        regions = policy.regions

        kc = mock.Mock()
        kc.validate_regions.return_value = regions.keys()
        policy._keystoneclient = kc

        self.patchobject(policy, '_create_plan', return_value=None)

        action = mock.Mock()
        action.action = 'CLUSTER_SCALE_OUT'
        action.context = self.context
        action.inputs = {}
        action.data = {'creation': {'count': 3}}

        cluster = mock.Mock()
        current_dist = {'R1': 0, 'R2': 0, 'R3': 0, 'R4': 0}
        cluster.get_region_distribution.return_value = current_dist
        mock_load.return_value = cluster

        res = policy.pre_op('FAKE_CLUSTER', action)

        self.assertIsNone(res)

        self.assertEqual('ERROR', action.data['status'])
        self.assertEqual('There is no feasible plan to handle all nodes.',
                         action.data['reason'])

        mock_load.assert_called_once_with(action.context, 'FAKE_CLUSTER')
        kc.validate_regions.assert_called_once_with(regions.keys())
        cluster.get_region_distribution.assert_called_once_with(regions.keys())
        policy._create_plan.assert_called_once_with(
            current_dist, regions, 3, True)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy

import mock
import six

from senlin.common import exception as exc
from senlin.profiles.os.heat import stack
from senlin.tests.unit.common import base
from senlin.tests.unit.common import utils


class TestHeatStackProfile(base.SenlinTestCase):

    def setUp(self):
        super(TestHeatStackProfile, self).setUp()
        self.context = utils.dummy_context()
        self.spec = {
            'type': 'os.heat.stack',
            'version': '1.0',
            'properties': {
                'template': {"Template": "data"},
                'template_url': '/test_uri',
                'context': {},
                'parameters': {'foo': 'bar'},
                'files': {},
                'timeout': 60,
                'disable_rollback': True,
                'environment': {}
            }
        }

    def test_stack_init(self):
        profile = stack.StackProfile('t', self.spec)
        self.assertIsNone(profile.stack_id)

    def test_do_validate(self):
        oc = mock.Mock()
        profile = stack.StackProfile('t', self.spec)
        profile._orchestrationclient = oc
        node_obj = mock.Mock(user='fake_user', project='fake_project')

        res = profile.do_validate(node_obj)

        props = self.spec['properties']
        call_args = {
            'stack_name': mock.ANY,
            'template': props['template'],
            'template_url': props['template_url'],
            'parameters': props['parameters'],
            'files': props['files'],
            'environment': props['environment'],
            'preview': True,
        }
        self.assertTrue(res)
        oc.stack_create.assert_called_once_with(**call_args)

    def test_do_validate_fails(self):
        oc = mock.Mock()
        profile = stack.StackProfile('t', self.spec)
        profile._orchestrationclient = oc
        err = exc.InternalError(code=400, message='Boom')
        oc.stack_create = mock.Mock(side_effect=err)
        node_obj = mock.Mock()
        node_obj.name = 'stack_node'

        ex = self.assertRaises(exc.InvalidSpec,
                               profile.do_validate, node_obj)

        props = self.spec['properties']
        call_args = {
            'stack_name': mock.ANY,
            'template': props['template'],
            'template_url': props['template_url'],
            'parameters': props['parameters'],
            'files': props['files'],
            'environment': props['environment'],
            'preview': True,
        }
        oc.stack_create.assert_called_once_with(**call_args)
        self.assertEqual('Failed in validating template: Boom',
                         six.text_type(ex))

    def test_do_create(self):
        oc = mock.Mock()
        profile = stack.StackProfile('t', self.spec)
        profile._orchestrationclient = oc
        node = mock.Mock(id='NODE_ID', cluster_id='CLUSTER_ID', index=123)
        node.name = 'test_node'
        fake_stack = mock.Mock(id='FAKE_ID')
        oc.stack_create = mock.Mock(return_value=fake_stack)

        # do it
        res = profile.do_create(node)

        # assertions
        kwargs = {
            'stack_name': mock.ANY,
            'template': self.spec['properties']['template'],
            'template_url': self.spec['properties']['template_url'],
            'timeout_mins': self.spec['properties']['timeout'],
            'disable_rollback': self.spec['properties']['disable_rollback'],
            'parameters': self.spec['properties']['parameters'],
            'files': self.spec['properties']['files'],
            'environment': self.spec['properties']['environment'],
            'tags': ",".join(['cluster_node_id=NODE_ID',
                              'cluster_id=CLUSTER_ID',
                              'cluster_node_index=123'])
        }
        self.assertEqual('FAKE_ID', res)
        oc.stack_create.assert_called_once_with(**kwargs)
        oc.wait_for_stack.assert_called_once_with('FAKE_ID', 'CREATE_COMPLETE',
                                                  timeout=3600)

    def test_do_create_with_template_url(self):
        spec = {
            'type': 'os.heat.stack',
            'version': '1.0',
            'properties': {
                'template': {},
                'template_url': '/test_uri',
                'context': {},
                'parameters': {'foo': 'bar'},
                'files': {},
                'timeout': 60,
                'disable_rollback': True,
                'environment': {}
            }
        }
        oc = mock.Mock()
        profile = stack.StackProfile('t', spec)
        profile._orchestrationclient = oc
        node = mock.Mock(id='NODE_ID', cluster_id='CLUSTER_ID', index=123)
        node.name = 'test_node'
        fake_stack = mock.Mock(id='FAKE_ID')
        oc.stack_create = mock.Mock(return_value=fake_stack)

        # do it
        res = profile.do_create(node)

        # assertions
        kwargs = {
            'stack_name': mock.ANY,
            'template': spec['properties']['template'],
            'template_url': spec['properties']['template_url'],
            'timeout_mins': spec['properties']['timeout'],
            'disable_rollback': spec['properties']['disable_rollback'],
            'parameters': spec['properties']['parameters'],
            'files': spec['properties']['files'],
            'environment': spec['properties']['environment'],
            'tags': ",".join(['cluster_node_id=NODE_ID',
                              'cluster_id=CLUSTER_ID',
                              'cluster_node_index=123'])
        }
        self.assertEqual('FAKE_ID', res)
        oc.stack_create.assert_called_once_with(**kwargs)
        oc.wait_for_stack.assert_called_once_with('FAKE_ID', 'CREATE_COMPLETE',
                                                  timeout=3600)

    def test_do_create_default_timeout(self):
        spec = copy.deepcopy(self.spec)
        del spec['properties']['timeout']
        profile = stack.StackProfile('t', spec)
        oc = mock.Mock()
        profile._orchestrationclient = oc
        node = mock.Mock(id='NODE_ID', cluster_id='CLUSTER_ID', index=123)
        node.name = 'test_node'
        fake_stack = mock.Mock(id='FAKE_ID')

        oc.stack_create = mock.Mock(return_value=fake_stack)
        oc.wait_for_stack = mock.Mock()

        # do it
        res = profile.do_create(node)

        # assertions
        self.assertEqual('FAKE_ID', res)
        kwargs = {
            'stack_name': mock.ANY,
            'template': self.spec['properties']['template'],
            'template_url': self.spec['properties']['template_url'],
            'timeout_mins': None,
            'disable_rollback': self.spec['properties']['disable_rollback'],
            'parameters': self.spec['properties']['parameters'],
            'files': self.spec['properties']['files'],
            'environment': self.spec['properties']['environment'],
            'tags': ",".join(['cluster_node_id=NODE_ID',
                              'cluster_id=CLUSTER_ID',
                              'cluster_node_index=123'])
        }
        oc.stack_create.assert_called_once_with(**kwargs)
        oc.wait_for_stack.assert_called_once_with('FAKE_ID', 'CREATE_COMPLETE',
                                                  timeout=None)

    def test_do_create_failed_create(self):
        oc = mock.Mock()
        profile = stack.StackProfile('t', self.spec)

        node = mock.Mock(id='NODE_ID', cluster_id='CLUSTER_ID', index=123)
        node.name = 'test_node'
        err = exc.InternalError(code=400, message='Too Bad')
        oc.stack_create = mock.Mock(side_effect=err)
        profile._orchestrationclient = oc

        # do it
        ex = self.assertRaises(exc.EResourceCreation,
                               profile.do_create,
                               node)

        # assertions
        self.assertEqual('Failed in creating stack: Too Bad.',
                         six.text_type(ex))
        call_args = {
            'stack_name': mock.ANY,
            'template': self.spec['properties']['template'],
            'template_url': self.spec['properties']['template_url'],
            'timeout_mins': self.spec['properties']['timeout'],
            'disable_rollback': self.spec['properties']['disable_rollback'],
            'parameters': self.spec['properties']['parameters'],
            'files': self.spec['properties']['files'],
            'environment': self.spec['properties']['environment'],
            'tags': ",".join(['cluster_node_id=NODE_ID',
                              'cluster_id=CLUSTER_ID',
                              'cluster_node_index=123'])
        }
        oc.stack_create.assert_called_once_with(**call_args)
        self.assertEqual(0, oc.wait_for_stack.call_count)

    def test_do_create_failed_wait(self):
        spec = copy.deepcopy(self.spec)
        del spec['properties']['timeout']
        profile = stack.StackProfile('t', spec)
        oc = mock.Mock()
        node = mock.Mock(id='NODE_ID', cluster_id='CLUSTER_ID', index=123)
        node.name = 'test_node'
        fake_stack = mock.Mock(id='FAKE_ID')

        oc.stack_create = mock.Mock(return_value=fake_stack)
        err = exc.InternalError(code=400, message='Timeout')
        oc.wait_for_stack = mock.Mock(side_effect=err)
        profile._orchestrationclient = oc

        # do it
        ex = self.assertRaises(exc.EResourceCreation,
                               profile.do_create,
                               node)

        # assertions
        self.assertEqual('Failed in creating stack: Timeout.',
                         six.text_type(ex))
        kwargs = {
            'stack_name': mock.ANY,
            'template': self.spec['properties']['template'],
            'template_url': self.spec['properties']['template_url'],
            'timeout_mins': None,
            'disable_rollback': self.spec['properties']['disable_rollback'],
            'parameters': self.spec['properties']['parameters'],
            'files': self.spec['properties']['files'],
            'environment': self.spec['properties']['environment'],
            'tags': ",".join(['cluster_node_id=NODE_ID',
                              'cluster_id=CLUSTER_ID',
                              'cluster_node_index=123'])
        }
        oc.stack_create.assert_called_once_with(**kwargs)
        oc.wait_for_stack.assert_called_once_with('FAKE_ID', 'CREATE_COMPLETE',
                                                  timeout=None)

    def test_do_delete(self):
        profile = stack.StackProfile('t', self.spec)
        oc = mock.Mock()
        profile._orchestrationclient = oc
        test_stack = mock.Mock(physical_id='FAKE_ID')

        # do it
        res = profile.do_delete(test_stack)

        # assertions
        self.assertTrue(res)
        oc.stack_delete.assert_called_once_with('FAKE_ID', True)
        oc.wait_for_stack_delete.assert_called_once_with('FAKE_ID')

    def test_do_delete_no_physical_id(self):
        profile = stack.StackProfile('t', self.spec)
        oc = mock.Mock()
        test_stack = mock.Mock(physical_id=None)
        profile._orchestrationclient = oc

        # do it
        res = profile.do_delete(test_stack, ignore_missing=False)

        # assertions
        self.assertTrue(res)
        self.assertFalse(oc.stack_delete.called)
        self.assertFalse(oc.wait_for_stack_delete.called)

    def test_do_delete_ignore_missing(self):
        profile = stack.StackProfile('t', self.spec)
        oc = mock.Mock()
        test_stack = mock.Mock(physical_id='FAKE_ID')
        profile._orchestrationclient = oc

        # do it
        res = profile.do_delete(test_stack, ignore_missing=False)

        # assertions
        self.assertTrue(res)
        oc.stack_delete.assert_called_once_with('FAKE_ID', False)
        oc.wait_for_stack_delete.assert_called_once_with('FAKE_ID')

    def test_do_delete_failed_deletion(self):
        profile = stack.StackProfile('t', self.spec)
        oc = mock.Mock()
        profile._orchestrationclient = oc
        err = exc.InternalError(code=400, message='Boom')
        oc.stack_delete = mock.Mock(side_effect=err)
        test_stack = mock.Mock(physical_id='FAKE_ID')

        # do it
        ex = self.assertRaises(exc.EResourceDeletion,
                               profile.do_delete,
                               test_stack)

        # assertions
        self.assertEqual("Failed in deleting stack 'FAKE_ID': Boom.",
                         six.text_type(ex))
        oc.stack_delete.assert_called_once_with('FAKE_ID', True)
        self.assertEqual(0, oc.wait_for_stack_delete.call_count)

    def test_do_delete_failed_timeout(self):
        profile = stack.StackProfile('t', self.spec)
        oc = mock.Mock()
        test_stack = mock.Mock(physical_id='FAKE_ID')
        profile._orchestrationclient = oc
        err = exc.InternalError(code=400, message='Boom')
        oc.wait_for_stack_delete = mock.Mock(side_effect=err)

        # do it
        ex = self.assertRaises(exc.EResourceDeletion,
                               profile.do_delete, test_stack)

        # assertions
        self.assertEqual("Failed in deleting stack 'FAKE_ID': Boom.",
                         six.text_type(ex))
        oc.stack_delete.assert_called_once_with('FAKE_ID', True)
        oc.wait_for_stack_delete.assert_called_once_with('FAKE_ID')

    def test_do_update(self):
        profile = stack.StackProfile('t', self.spec)
        oc = mock.Mock()
        profile._orchestrationclient = oc
        test_stack = mock.Mock(physical_id='FAKE_ID')
        new_spec = {
            'type': 'os.heat.stack',
            'version': '1.0',
            'properties': {
                'template': {"Template": "data update"},
                'context': {},
                'parameters': {'new': 'params'},
                'files': {'file1': 'new_content'},
                'timeout': 123,
                'disable_rollback': False,
                'environment': {'foo': 'bar'}
            }
        }
        new_profile = stack.StackProfile('u', new_spec)

        # do it
        res = profile.do_update(test_stack, new_profile)

        # assertions
        self.assertTrue(res)
        kwargs = {
            'template': {'Template': 'data update'},
            'parameters': {'new': 'params'},
            'timeout_mins': 123,
            'disable_rollback': False,
            'files': {'file1': 'new_content'},
            'environment': {'foo': 'bar'},
        }
        oc.stack_update.assert_called_once_with('FAKE_ID', **kwargs)
        oc.wait_for_stack.assert_called_once_with(
            'FAKE_ID', 'UPDATE_COMPLETE', timeout=3600)

    def test_do_update_no_physical_stack(self):
        profile = stack.StackProfile('t', self.spec)
        test_stack = mock.Mock(physical_id=None)
        new_profile = mock.Mock()

        res = profile.do_update(test_stack, new_profile)

        self.assertFalse(res)

    def test_do_update_only_template(self):
        profile = stack.StackProfile('t', self.spec)
        oc = mock.Mock()
        profile._orchestrationclient = oc
        stack_obj = mock.Mock(physical_id='FAKE_ID')
        new_spec = copy.deepcopy(self.spec)
        new_spec['properties']['template'] = {"Template": "data update"}
        new_profile = stack.StackProfile('u', new_spec)

        res = profile.do_update(stack_obj, new_profile)

        self.assertTrue(res)
        oc.stack_update.assert_called_once_with(
            'FAKE_ID', template={"Template": "data update"})
        oc.wait_for_stack.assert_called_once_with(
            'FAKE_ID', 'UPDATE_COMPLETE', timeout=3600)

    def test_do_update_only_params(self):
        profile = stack.StackProfile('t', self.spec)
        oc = mock.Mock()
        profile._orchestrationclient = oc
        stack_obj = mock.Mock(physical_id='FAKE_ID')
        new_spec = copy.deepcopy(self.spec)
        new_spec['properties']['parameters'] = {"new": "params"}
        new_profile = stack.StackProfile('u', new_spec)

        res = profile.do_update(stack_obj, new_profile)

        self.assertTrue(res)
        oc.stack_update.assert_called_once_with(
            'FAKE_ID', parameters={"new": "params"})
        oc.wait_for_stack.assert_called_once_with(
            'FAKE_ID', 'UPDATE_COMPLETE', timeout=3600)

    def test_do_update_with_timeout_value(self):
        profile = stack.StackProfile('t', self.spec)
        oc = mock.Mock()
        profile._orchestrationclient = oc
        stack_obj = mock.Mock(physical_id='FAKE_ID')
        new_spec = copy.deepcopy(self.spec)
        new_spec['properties']['timeout'] = 120
        new_profile = stack.StackProfile('u', new_spec)

        # do it
        res = profile.do_update(stack_obj, new_profile)

        # assertions
        self.assertTrue(res)
        oc.stack_update.assert_called_once_with('FAKE_ID', timeout_mins=120)
        oc.wait_for_stack.assert_called_once_with(
            'FAKE_ID', 'UPDATE_COMPLETE', timeout=3600)

    def test_do_update_disable_rollback(self):
        profile = stack.StackProfile('t', self.spec)
        oc = mock.Mock()
        profile._orchestrationclient = oc
        stack_obj = mock.Mock(physical_id='FAKE_ID')
        new_spec = copy.deepcopy(self.spec)
        new_spec['properties']['disable_rollback'] = False
        new_profile = stack.StackProfile('u', new_spec)

        # do it
        res = profile.do_update(stack_obj, new_profile)

        # assertions
        self.assertTrue(res)
        oc.stack_update.assert_called_once_with('FAKE_ID',
                                                disable_rollback=False)
        oc.wait_for_stack.assert_called_once_with('FAKE_ID', 'UPDATE_COMPLETE',
                                                  timeout=3600)

    def test_do_update_files(self):
        profile = stack.StackProfile('t', self.spec)
        oc = mock.Mock()
        profile._orchestrationclient = oc
        stack_obj = mock.Mock(physical_id='FAKE_ID')
        new_spec = copy.deepcopy(self.spec)
        new_spec['properties']['files'] = {"new": "file1"}
        new_profile = stack.StackProfile('u', new_spec)

        # do it
        res = profile.do_update(stack_obj, new_profile)

        # assertions
        self.assertTrue(res)
        oc.stack_update.assert_called_once_with(
            'FAKE_ID', files={"new": "file1"})
        oc.wait_for_stack.assert_called_once_with(
            'FAKE_ID', 'UPDATE_COMPLETE', timeout=3600)

    def test_do_update_environment(self):
        profile = stack.StackProfile('t', self.spec)
        oc = mock.Mock()
        profile._orchestrationclient = oc
        stack_obj = mock.Mock(physical_id='FAKE_ID')
        new_spec = copy.deepcopy(self.spec)
        new_spec['properties']['environment'] = {"new": "env1"}
        new_profile = stack.StackProfile('u', new_spec)

        # do it
        res = profile.do_update(stack_obj, new_profile)

        # assertions
        self.assertTrue(res)
        oc.stack_update.assert_called_once_with(
            'FAKE_ID', environment={"new": "env1"})
        oc.wait_for_stack.assert_called_once_with(
            'FAKE_ID', 'UPDATE_COMPLETE', timeout=3600)

    def test_do_update_no_change(self):
        profile = stack.StackProfile('t', self.spec)
        oc = mock.Mock()
        profile._orchestrationclient = oc
        stack_obj = mock.Mock(physical_id='FAKE_ID')
        new_spec = copy.deepcopy(self.spec)
        new_profile = stack.StackProfile('u', new_spec)

        res = profile.do_update(stack_obj, new_profile)

        self.assertTrue(res)
        self.assertEqual(0, oc.stack_update.call_count)

    def test_do_update_failed_update(self):
        profile = stack.StackProfile('t', self.spec)
        oc = mock.Mock()
        profile._orchestrationclient = oc
        oc.stack_update = mock.Mock(
            side_effect=exc.InternalError(code=400, message='Failed'))
        stack_obj = mock.Mock(physical_id='FAKE_ID')
        new_spec = copy.deepcopy(self.spec)
        new_spec['properties']['environment'] = {"new": "env1"}
        new_profile = stack.StackProfile('u', new_spec)

        ex = self.assertRaises(exc.EResourceUpdate,
                               profile.do_update,
                               stack_obj, new_profile)

        oc.stack_update.assert_called_once_with(
            'FAKE_ID', environment={"new": "env1"})
        self.assertEqual(0, oc.wait_for_stack.call_count)
        self.assertEqual("Failed in updating stack 'FAKE_ID': "
                         "Failed.", six.text_type(ex))

    def test_do_update_timeout(self):
        profile = stack.StackProfile('t', self.spec)
        oc = mock.Mock()
        profile._orchestrationclient = oc
        oc.wait_for_stack = mock.Mock(
            side_effect=exc.InternalError(code=400, message='Timeout'))
        stack_obj = mock.Mock(physical_id='FAKE_ID')
        new_spec = copy.deepcopy(self.spec)
        new_spec['properties']['environment'] = {"new": "env1"}
        new_profile = stack.StackProfile('u', new_spec)

        ex = self.assertRaises(exc.EResourceUpdate,
                               profile.do_update,
                               stack_obj, new_profile)

        oc.stack_update.assert_called_once_with(
            'FAKE_ID', environment={"new": "env1"})
        oc.wait_for_stack.assert_called_once_with(
            'FAKE_ID', 'UPDATE_COMPLETE', timeout=3600)
        self.assertEqual("Failed in updating stack 'FAKE_ID': "
                         "Timeout.", six.text_type(ex))

    def test_do_check(self):
        node_obj = mock.Mock(physical_id='FAKE_ID')
        profile = stack.StackProfile('t', self.spec)
        oc = mock.Mock()
        profile._orchestrationclient = oc

        # do it
        res = profile.do_check(node_obj)

        # assertions
        self.assertTrue(res)
        oc.stack_check.assert_called_once_with('FAKE_ID')
        oc.wait_for_stack.assert_called_once_with(
            'FAKE_ID', 'CHECK_COMPLETE', timeout=3600)

    def test_do_check_no_physical_id(self):
        node_obj = mock.Mock(physical_id=None)
        profile = stack.StackProfile('t', self.spec)

        res = profile.do_check(node_obj)

        self.assertFalse(res)

    def test_do_check_failed_checking(self):
        node_obj = mock.Mock(physical_id='FAKE_ID')
        profile = stack.StackProfile('t', self.spec)
        oc = mock.Mock()
        profile._orchestrationclient = oc
        oc.stack_check = mock.Mock(
            side_effect=exc.InternalError(code=400, message='BOOM'))

        self.assertRaises(exc.EResourceOperation, profile.do_check, node_obj)

        oc.stack_check.assert_called_once_with('FAKE_ID')
        self.assertEqual(0, oc.wait_for_stack.call_count)

    def test_do_check_failed_in_waiting(self):
        node_obj = mock.Mock(physical_id='FAKE_ID')
        profile = stack.StackProfile('t', self.spec)
        oc = mock.Mock()
        profile._orchestrationclient = oc
        oc.wait_for_stack = mock.Mock(
            side_effect=exc.InternalError(code=400, message='BOOM'))

        self.assertRaises(exc.EResourceOperation, profile.do_check, node_obj)

        oc.stack_check.assert_called_once_with('FAKE_ID')
        oc.wait_for_stack.assert_called_once_with(
            'FAKE_ID', 'CHECK_COMPLETE', timeout=3600)

    def test_do_get_details(self):
        profile = stack.StackProfile('t', self.spec)
        oc = mock.Mock()
        profile._orchestrationclient = oc
        details = mock.Mock()
        details.to_dict.return_value = {'foo': 'bar'}
        oc.stack_get = mock.Mock(return_value=details)
        node_obj = mock.Mock(physical_id='FAKE_ID')

        res = profile.do_get_details(node_obj)

        self.assertEqual({'foo': 'bar'}, res)
        oc.stack_get.assert_called_once_with('FAKE_ID')

    def test_do_get_details_no_physical_id(self):
        profile = stack.StackProfile('t', self.spec)
        node_obj = mock.Mock(physical_id=None)

        res = profile.do_get_details(node_obj)

        self.assertEqual({}, res)

    def test_do_get_details_failed_retrieval(self):
        profile = stack.StackProfile('t', self.spec)
        node_obj = mock.Mock(physical_id='STACK_ID')
        oc = mock.Mock()
        oc.stack_get.side_effect = exc.InternalError(message='BOOM')
        profile._orchestrationclient = oc

        res = profile.do_get_details(node_obj)

        self.assertEqual({'Error': {'code': 500, 'message': 'BOOM'}}, res)
        oc.stack_get.assert_called_once_with('STACK_ID')

    def test_do_adopt(self):
        profile = stack.StackProfile('t', self.spec)
        x_stack = mock.Mock(
            parameters={'p1': 'v1', 'OS::stack_id': 'FAKE_ID'},
            timeout_mins=123,
            is_rollback_disabled=False
        )
        oc = mock.Mock()
        oc.stack_get = mock.Mock(return_value=x_stack)

        # mock template
        templ = mock.Mock()
        templ.to_dict.return_value = {'foo': 'bar'}
        oc.stack_get_template = mock.Mock(return_value=templ)

        # mock environment
        env = mock.Mock()
        env.to_dict.return_value = {'ke': 've'}
        oc.stack_get_environment = mock.Mock(return_value=env)

        oc.stack_get_files = mock.Mock(return_value={'fn': 'content'})
        profile._orchestrationclient = oc

        node_obj = mock.Mock(physical_id='FAKE_ID')

        res = profile.do_adopt(node_obj)

        expected = {
            'environment': {'ke': 've'},
            'files': {'fn': 'content'},
            'template': {'foo': 'bar'},
            'parameters': {'p1': 'v1'},
            'timeout': 123,
            'disable_rollback': False
        }
        self.assertEqual(expected, res)
        oc.stack_get.assert_called_once_with('FAKE_ID')
        oc.stack_get_template.assert_called_once_with('FAKE_ID')
        oc.stack_get_environment.assert_called_once_with('FAKE_ID')
        oc.stack_get_files.assert_called_once_with('FAKE_ID')

    def test_do_adopt_failed_get(self):
        profile = stack.StackProfile('t', self.spec)
        oc = mock.Mock()
        oc.stack_get.side_effect = exc.InternalError(message='BOOM')
        profile._orchestrationclient = oc
        node_obj = mock.Mock(physical_id='FAKE_ID')

        res = profile.do_adopt(node_obj)

        expected = {'Error': {'code': 500, 'message': 'BOOM'}}
        self.assertEqual(expected, res)
        oc.stack_get.assert_called_once_with('FAKE_ID')

    def test_do_adopt_failed_get_template(self):
        profile = stack.StackProfile('t', self.spec)
        x_stack = mock.Mock()
        oc = mock.Mock()
        oc.stack_get = mock.Mock(return_value=x_stack)
        oc.stack_get_template.side_effect = exc.InternalError(message='BOOM')
        profile._orchestrationclient = oc

        node_obj = mock.Mock(physical_id='FAKE_ID')

        res = profile.do_adopt(node_obj)

        expected = {'Error': {'code': 500, 'message': 'BOOM'}}
        self.assertEqual(expected, res)
        oc.stack_get.assert_called_once_with('FAKE_ID')
        oc.stack_get_template.assert_called_once_with('FAKE_ID')

    def test_do_adopt_failed_get_environment(self):
        profile = stack.StackProfile('t', self.spec)
        x_stack = mock.Mock()
        oc = mock.Mock()
        oc.stack_get = mock.Mock(return_value=x_stack)
        oc.stack_get_template = mock.Mock(return_value={'foo': 'bar'})
        err = exc.InternalError(message='BOOM')
        oc.stack_get_environment.side_effect = err
        profile._orchestrationclient = oc

        node_obj = mock.Mock(physical_id='FAKE_ID')

        res = profile.do_adopt(node_obj)

        expected = {'Error': {'code': 500, 'message': 'BOOM'}}
        self.assertEqual(expected, res)
        oc.stack_get.assert_called_once_with('FAKE_ID')
        oc.stack_get_template.assert_called_once_with('FAKE_ID')
        oc.stack_get_environment.assert_called_once_with('FAKE_ID')

    def test_do_adopt_failed_get_files(self):
        profile = stack.StackProfile('t', self.spec)
        x_stack = mock.Mock()
        oc = mock.Mock()
        oc.stack_get = mock.Mock(return_value=x_stack)
        oc.stack_get_template = mock.Mock(return_value={'foo': 'bar'})
        oc.stack_get_environment = mock.Mock(return_value={'ke': 've'})
        oc.stack_get_files.side_effect = exc.InternalError(message='BOOM')
        profile._orchestrationclient = oc

        node_obj = mock.Mock(physical_id='FAKE_ID')

        res = profile.do_adopt(node_obj)

        expected = {'Error': {'code': 500, 'message': 'BOOM'}}
        self.assertEqual(expected, res)
        oc.stack_get.assert_called_once_with('FAKE_ID')
        oc.stack_get_template.assert_called_once_with('FAKE_ID')
        oc.stack_get_environment.assert_called_once_with('FAKE_ID')
        oc.stack_get_files.assert_called_once_with('FAKE_ID')

    def test_do_adopt_with_overrides(self):
        profile = stack.StackProfile('t', self.spec)
        x_stack = mock.Mock(
            parameters={'p1': 'v1', 'OS::stack_id': 'FAKE_ID'},
            timeout_mins=123,
            is_rollback_disabled=False
        )
        oc = mock.Mock()
        oc.stack_get = mock.Mock(return_value=x_stack)

        # mock environment
        env = mock.Mock()
        env.to_dict.return_value = {'ke': 've'}
        oc.stack_get_environment = mock.Mock(return_value=env)

        # mock template
        templ = mock.Mock()
        templ.to_dict.return_value = {'foo': 'bar'}
        oc.stack_get_template = mock.Mock(return_value=templ)

        oc.stack_get_files = mock.Mock(return_value={'fn': 'content'})
        profile._orchestrationclient = oc

        node_obj = mock.Mock(physical_id='FAKE_ID')
        overrides = {'environment': {'ENV': 'SETTING'}}
        res = profile.do_adopt(node_obj, overrides=overrides)

        expected = {
            'environment': {'ENV': 'SETTING'},
            'files': {'fn': 'content'},
            'template': {'foo': 'bar'},
            'parameters': {'p1': 'v1'},
            'timeout': 123,
            'disable_rollback': False
        }
        self.assertEqual(expected, res)
        oc.stack_get.assert_called_once_with('FAKE_ID')
        oc.stack_get_template.assert_called_once_with('FAKE_ID')

    def test__refresh_tags_empty_no_add(self):
        profile = stack.StackProfile('t', self.spec)
        node = mock.Mock()

        res = profile._refresh_tags([], node, False)

        self.assertEqual(("", False), res)

    def test__refresh_tags_with_contents_no_add(self):
        profile = stack.StackProfile('t', self.spec)
        node = mock.Mock()

        res = profile._refresh_tags(['foo'], node, False)

        self.assertEqual(('foo', False), res)

    def test__refresh_tags_deleted_no_add(self):
        profile = stack.StackProfile('t', self.spec)
        node = mock.Mock()

        res = profile._refresh_tags(['cluster_id=FOO', 'bar'], node, False)

        self.assertEqual(('bar', True), res)

    def test__refresh_tags_empty_and_add(self):
        profile = stack.StackProfile('t', self.spec)
        node = mock.Mock(id='NODE_ID', cluster_id='CLUSTER_ID', index=123)

        res = profile._refresh_tags([], node, True)

        expected = ",".join(['cluster_id=CLUSTER_ID',
                             'cluster_node_id=NODE_ID',
                             'cluster_node_index=123'])
        self.assertEqual((expected, True), res)

    def test__refresh_tags_with_contents_and_add(self):
        profile = stack.StackProfile('t', self.spec)
        node = mock.Mock(id='NODE_ID', cluster_id='CLUSTER_ID', index=123)

        res = profile._refresh_tags(['foo'], node, True)

        expected = ",".join(['foo',
                             'cluster_id=CLUSTER_ID',
                             'cluster_node_id=NODE_ID',
                             'cluster_node_index=123'])
        self.assertEqual((expected, True), res)

    def test__refresh_tags_deleted_and_add(self):
        profile = stack.StackProfile('t', self.spec)
        node = mock.Mock(id='NODE_ID', cluster_id='CLUSTER_ID', index=123)

        res = profile._refresh_tags(['cluster_id=FOO', 'bar'], node, True)

        expected = ",".join(['bar',
                             'cluster_id=CLUSTER_ID',
                             'cluster_node_id=NODE_ID',
                             'cluster_node_index=123'])
        self.assertEqual((expected, True), res)

    def test_do_join(self):
        profile = stack.StackProfile('t', self.spec)
        oc = mock.Mock()
        profile._orchestrationclient = oc
        x_stack = mock.Mock(tags='foo')
        oc.stack_get.return_value = x_stack
        node = mock.Mock(physical_id='STACK_ID')
        mock_tags = self.patchobject(profile, '_refresh_tags',
                                     return_value=('bar', True))

        res = profile.do_join(node, 'CLUSTER_ID')

        self.assertTrue(res)
        oc.stack_get.assert_called_once_with('STACK_ID')
        mock_tags.assert_called_once_with('foo', node, True)
        oc.stack_update.assert_called_once_with('STACK_ID', **{'tags': 'bar'})

    def test_do_join_no_physical_id(self):
        profile = stack.StackProfile('t', self.spec)
        node = mock.Mock(physical_id=None)

        res = profile.do_join(node, 'CLUSTER_ID')

        self.assertFalse(res)

    def test_do_join_failed_get_stack(self):
        profile = stack.StackProfile('t', self.spec)
        oc = mock.Mock()
        profile._orchestrationclient = oc
        err = exc.InternalError(code=400, message='Boom')
        oc.stack_get.side_effect = err
        node = mock.Mock(physical_id='STACK_ID')

        res = profile.do_join(node, 'CLUSTER_ID')

        self.assertFalse(res)
        oc.stack_get.assert_called_once_with('STACK_ID')

    def test_do_join_no_update(self):
        profile = stack.StackProfile('t', self.spec)
        oc = mock.Mock()
        profile._orchestrationclient = oc
        x_stack = mock.Mock(tags='foo')
        oc.stack_get.return_value = x_stack
        node = mock.Mock(physical_id='STACK_ID')
        mock_tags = self.patchobject(profile, '_refresh_tags',
                                     return_value=('foo', False))

        res = profile.do_join(node, 'CLUSTER_ID')

        self.assertTrue(res)
        oc.stack_get.assert_called_once_with('STACK_ID')
        mock_tags.assert_called_once_with('foo', node, True)
        self.assertEqual(0, oc.stack_update.call_count)

    def test_do_join_failed_update(self):
        profile = stack.StackProfile('t', self.spec)
        oc = mock.Mock()
        profile._orchestrationclient = oc
        x_stack = mock.Mock(tags='foo')
        oc.stack_get.return_value = x_stack
        err = exc.InternalError(code=400, message='Boom')
        oc.stack_update.side_effect = err
        node = mock.Mock(physical_id='STACK_ID')
        mock_tags = self.patchobject(profile, '_refresh_tags',
                                     return_value=('bar', True))

        res = profile.do_join(node, 'CLUSTER_ID')

        self.assertFalse(res)
        oc.stack_get.assert_called_once_with('STACK_ID')
        mock_tags.assert_called_once_with('foo', node, True)
        oc.stack_update.assert_called_once_with('STACK_ID', **{'tags': 'bar'})

    def test_do_leave(self):
        profile = stack.StackProfile('t', self.spec)
        oc = mock.Mock()
        profile._orchestrationclient = oc
        x_stack = mock.Mock(tags='foo')
        oc.stack_get.return_value = x_stack
        node = mock.Mock(physical_id='STACK_ID')
        mock_tags = self.patchobject(profile, '_refresh_tags',
                                     return_value=('bar', True))

        res = profile.do_leave(node)

        self.assertTrue(res)
        oc.stack_get.assert_called_once_with('STACK_ID')
        mock_tags.assert_called_once_with('foo', node, False)
        oc.stack_update.assert_called_once_with('STACK_ID', **{'tags': 'bar'})

    def test_do_leave_no_physical_id(self):
        profile = stack.StackProfile('t', self.spec)
        node = mock.Mock(physical_id=None)

        res = profile.do_leave(node)

        self.assertFalse(res)

    def test_do_leave_failed_get_stack(self):
        profile = stack.StackProfile('t', self.spec)
        oc = mock.Mock()
        profile._orchestrationclient = oc
        err = exc.InternalError(code=400, message='Boom')
        oc.stack_get.side_effect = err
        node = mock.Mock(physical_id='STACK_ID')

        res = profile.do_leave(node)

        self.assertFalse(res)
        oc.stack_get.assert_called_once_with('STACK_ID')

    def test_do_leave_no_update(self):
        profile = stack.StackProfile('t', self.spec)
        oc = mock.Mock()
        profile._orchestrationclient = oc
        x_stack = mock.Mock(tags='foo')
        oc.stack_get.return_value = x_stack
        node = mock.Mock(physical_id='STACK_ID')
        mock_tags = self.patchobject(profile, '_refresh_tags',
                                     return_value=('foo', False))

        res = profile.do_leave(node)

        self.assertTrue(res)
        oc.stack_get.assert_called_once_with('STACK_ID')
        mock_tags.assert_called_once_with('foo', node, False)
        self.assertEqual(0, oc.stack_update.call_count)

    def test_do_leave_failed_update(self):
        profile = stack.StackProfile('t', self.spec)
        oc = mock.Mock()
        profile._orchestrationclient = oc
        x_stack = mock.Mock(tags='foo')
        oc.stack_get.return_value = x_stack
        err = exc.InternalError(code=400, message='Boom')
        oc.stack_update.side_effect = err
        node = mock.Mock(physical_id='STACK_ID')
        mock_tags = self.patchobject(profile, '_refresh_tags',
                                     return_value=('bar', True))

        res = profile.do_leave(node)

        self.assertFalse(res)
        oc.stack_get.assert_called_once_with('STACK_ID')
        mock_tags.assert_called_once_with('foo', node, False)
        oc.stack_update.assert_called_once_with('STACK_ID', **{'tags': 'bar'})



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import mock
import six
from webob import exc

from senlin.api.common import util
from senlin.api.middleware import fault
from senlin.api.openstack.v1 import policy_types
from senlin.common import exception as senlin_exc
from senlin.common import policy
from senlin.rpc import client as rpc_client
from senlin.tests.unit.api import shared
from senlin.tests.unit.common import base


@mock.patch.object(policy, 'enforce')
class PolicyTypeControllerTest(shared.ControllerTest, base.SenlinTestCase):
    def setUp(self):
        super(PolicyTypeControllerTest, self).setUp()

        class DummyConfig(object):
            bind_port = 8778

        cfgopts = DummyConfig()
        self.controller = policy_types.PolicyTypeController(options=cfgopts)

    @mock.patch.object(util, 'parse_request')
    @mock.patch.object(rpc_client.EngineClient, 'call')
    def test_list(self, mock_call, mock_parse, mock_enforce):
        self._mock_enforce_setup(mock_enforce, 'index', True)
        req = self._get('/policy_types')

        engine_response = [
            {'name': 'senlin.policy.p1', 'version': '1.0', 'attr': 'v1'},
            {'name': 'senlin.policy.p2', 'version': '1.0', 'attr': 'v2'}
        ]

        mock_call.return_value = engine_response
        obj = mock.Mock()
        mock_parse.return_value = obj

        response = self.controller.index(req)

        self.assertEqual(
            [
                {'name': 'senlin.policy.p1-1.0'},
                {'name': 'senlin.policy.p2-1.0'},
            ],
            response['policy_types']
        )
        mock_parse.assert_called_once_with(
            'PolicyTypeListRequest', req, {})
        mock_call.assert_called_once_with(
            req.context, 'policy_type_list', mock.ANY)

    @mock.patch.object(util, 'parse_request')
    @mock.patch.object(rpc_client.EngineClient, 'call')
    def test_list_old_version(self, mock_call, mock_parse, mock_enforce):
        self._mock_enforce_setup(mock_enforce, 'index', True)
        req = self._get('/policy_types', version='1.3')

        engine_response = [
            {'name': 'senlin.policy.p1', 'version': '1.0'},
            {'name': 'senlin.policy.p2', 'version': '1.1'}
        ]

        mock_call.return_value = engine_response
        obj = mock.Mock()
        mock_parse.return_value = obj

        response = self.controller.index(req)

        self.assertEqual(
            [
                {'name': 'senlin.policy.p1-1.0'},
                {'name': 'senlin.policy.p2-1.1'}
            ],
            response['policy_types']
        )
        mock_parse.assert_called_once_with(
            'PolicyTypeListRequest', req, {})
        mock_call.assert_called_once_with(
            req.context, 'policy_type_list', mock.ANY)

    @mock.patch.object(util, 'parse_request')
    @mock.patch.object(rpc_client.EngineClient, 'call')
    def test_list_new_version(self, mock_call, mock_parse, mock_enforce):
        self._mock_enforce_setup(mock_enforce, 'index', True)
        req = self._get('/policy_types', version='1.5')

        engine_response = [
            {'name': 'senlin.policy.p1', 'version': '1.0', 'a1': 'v1'},
            {'name': 'senlin.policy.p2', 'version': '1.1', 'a2': 'v2'}
        ]

        mock_call.return_value = engine_response
        obj = mock.Mock()
        mock_parse.return_value = obj

        response = self.controller.index(req)

        self.assertEqual(engine_response, response['policy_types'])
        mock_parse.assert_called_once_with(
            'PolicyTypeListRequest', req, {})
        mock_call.assert_called_once_with(
            req.context, 'policy_type_list', mock.ANY)

    def test_list_err_denied_policy(self, mock_enforce):
        self._mock_enforce_setup(mock_enforce, 'index', False)
        req = self._get('/policy_types')
        resp = shared.request_with_middleware(fault.FaultWrapper,
                                              self.controller.index,
                                              req)

        self.assertEqual(403, resp.status_int)
        self.assertIn('403 Forbidden', six.text_type(resp))

    @mock.patch.object(util, 'parse_request')
    @mock.patch.object(rpc_client.EngineClient, 'call')
    def test_get_old_version(self, mock_call, mock_parse, mock_enforce):
        self._mock_enforce_setup(mock_enforce, 'get', True)
        type_name = 'SimplePolicy'
        req = self._get('/policy_types/%(type)s' % {'type': type_name},
                        version='1.3')

        engine_response = {
            'name': type_name,
            'schema': {
                'Foo': {'type': 'String', 'required': False},
                'Bar': {'type': 'Integer', 'required': False},
            },
        }

        mock_call.return_value = engine_response
        obj = mock.Mock()
        mock_parse.return_value = obj

        response = self.controller.get(req, type_name=type_name)

        self.assertEqual(engine_response, response['policy_type'])
        mock_parse.assert_called_once_with(
            'PolicyTypeGetRequest', req, {'type_name': type_name})
        mock_call.assert_called_once_with(
            req.context, 'policy_type_get', mock.ANY)

    @mock.patch.object(util, 'parse_request')
    @mock.patch.object(rpc_client.EngineClient, 'call')
    def test_get_new_version(self, mock_call, mock_parse, mock_enforce):
        self._mock_enforce_setup(mock_enforce, 'get', True)
        type_name = 'SimplePolicy'
        req = self._get('/policy_types/%(type)s' % {'type': type_name},
                        version='1.5')

        engine_response = {
            'name': type_name,
            'schema': {
                'Foo': {'type': 'String', 'required': False},
                'Bar': {'type': 'Integer', 'required': False},
            },
            'support_status': 'faked_status'
        }

        mock_call.return_value = engine_response
        obj = mock.Mock()
        mock_parse.return_value = obj

        response = self.controller.get(req, type_name=type_name)

        self.assertEqual(engine_response, response['policy_type'])
        mock_parse.assert_called_once_with(
            'PolicyTypeGetRequest', req, {'type_name': type_name})
        mock_call.assert_called_once_with(
            req.context, 'policy_type_get', mock.ANY)

    @mock.patch.object(util, 'parse_request')
    @mock.patch.object(rpc_client.EngineClient, 'call')
    def test_policy_type_get(self, mock_call, mock_parse, mock_enforce):
        self._mock_enforce_setup(mock_enforce, 'get', True)
        type_name = 'SimplePolicy'
        req = self._get('/policy_types/%(type)s' % {'type': type_name})

        engine_response = {
            'name': type_name,
            'schema': {
                'Foo': {'type': 'String', 'required': False},
                'Bar': {'type': 'Integer', 'required': False},
            },
        }

        mock_call.return_value = engine_response
        obj = mock.Mock()
        mock_parse.return_value = obj

        response = self.controller.get(req, type_name=type_name)

        self.assertEqual(engine_response, response['policy_type'])
        mock_parse.assert_called_once_with(
            'PolicyTypeGetRequest', req, {'type_name': type_name})
        mock_call.assert_called_once_with(
            req.context, 'policy_type_get', mock.ANY)

    @mock.patch.object(util, 'parse_request')
    @mock.patch.object(rpc_client.EngineClient, 'call')
    def test_policy_type_get_not_found(self, mock_call, mock_parse,
                                       mock_enforce):
        self._mock_enforce_setup(mock_enforce, 'get', True)
        type_name = 'BogusPolicyType'
        req = self._get('/policy_types/%(type)s' % {'type': type_name})

        error = senlin_exc.ResourceNotFound(type='policy_type', id=type_name)
        mock_call.side_effect = shared.to_remote_error(error)

        resp = shared.request_with_middleware(fault.FaultWrapper,
                                              self.controller.get,
                                              req, type_name=type_name)
        self.assertEqual(404, resp.json['code'])
        self.assertEqual('ResourceNotFound', resp.json['error']['type'])

    @mock.patch.object(util, 'parse_request')
    @mock.patch.object(rpc_client.EngineClient, 'call')
    def test_policy_type_get_bad_param(self, mock_call, mock_parse,
                                       mock_enforce):
        self._mock_enforce_setup(mock_enforce, 'get', True)
        type_name = 11
        req = self._get('/policy_types/%(type)s' % {'type': type_name})

        mock_parse.side_effect = exc.HTTPBadRequest("bad param")
        ex = self.assertRaises(exc.HTTPBadRequest,
                               self.controller.get,
                               req, type_name=type_name)
        self.assertEqual("bad param", six.text_type(ex))
        mock_parse.assert_called_once_with(
            'PolicyTypeGetRequest', req, {'type_name': type_name})
        self.assertEqual(0, mock_call.call_count)

    def test_policy_type_schema_err_denied_policy(self, mock_enforce):
        self._mock_enforce_setup(mock_enforce, 'get', False)
        type_name = 'FakePolicyType'
        req = self._get('/policy_types/%(type)s' % {'type': type_name})

        resp = shared.request_with_middleware(fault.FaultWrapper,
                                              self.controller.get,
                                              req, type_name=type_name)
        self.assertEqual(403, resp.status_int)
        self.assertIn('403 Forbidden', six.text_type(resp))
